from pyautogui import *
import time

WAIT_AFTER_SEC_FLOAT = 0.4 # Настройка модуля Mouse: Время, которое ожидать после выполнения любой операции модуля Mouse. Настройка является единой для всех участов кода, использующих модуль Mouse. Если для некоторой функции требуется изменить данное время ожидания, то в отношении этой функции можно применить соответсвующий аргумент.

def Click(inXInt:int=None, inYInt:int=None, inClickCountInt:int=1, inIntervalSecFloat:float=0.0, inButtonStr:str='left', inMoveDurationSecFloat:float=0.0, inWaitAfterSecFloat:float=WAIT_AFTER_SEC_FLOAT):
    """Нажатие (вниз) кнопки мыши и затем немедленно выпуск (вверх) её. Допускается следующая параметризация. Если не указаны inXInt и inYInt - клик производится по месту нахождения указателя мыши.

    !ВНИМАНИЕ! Отсчет координат inXInt, inYInt начинается с левого верхнего края рабочей области (экрана).
    
    .. code-block:: python

        # Mouse: Взаимодействие с мышью
        from pyOpenRPA.Robot import Mouse
        Mouse.Click(100,150) #Выполнить нажатие левой клавиши мыши на экране по координатам: X(гор) 100px, Y(вер) 150px.

    :param inXInt: Целевая позиция указателя мыши по оси X (горизонтальная ось). 
    :type inXInt: int, опциональный
    :param inYInt: Целевая позиция указателя мыши по оси Y (вертикальная ось). 
    :type inYInt: int, опциональный
    :param inClickCountInt: Количество нажатий (вниз и вверх) кнопкой мыши, По умолчанию 1
    :type inClickCountInt: int, опциональный
    :param inIntervalSecFloat: Интервал ожидания в секундах между нажатиями, По умолчанию 0.0
    :type inIntervalSecFloat: float, опциональный 
    :param inButtonStr: Номер кнопки, которую требуется нажать. Возможные варианты: 'left', 'middle', 'right' или 1, 2, 3. В остальных случаях инициирует исключение ValueError. По умолчанию 'left'
    :type inButtonStr: str, опциональный
    :param inMoveDurationSecFloat: Время перемещения указателя мыши, По умолчанию 0.0 (моментальное перемещение)
    :type inMoveDurationSecFloat: float, опциональный
    :param inWaitAfterSecFloat: Количество секунд, которые ожидать после выполнения операции. По умолчанию установлено в настройках модуля Mouse (базовое значение 0.4)
    :type inWaitAfterSecFloat: float, опциональный
    """
    click(x=inXInt, y=inYInt, clicks=inClickCountInt, interval=inIntervalSecFloat, button=inButtonStr, duration=inMoveDurationSecFloat)
    time.sleep(inWaitAfterSecFloat)

def ClickDouble(inXInt:int=None, inYInt:int=None, inWaitAfterSecFloat:float=WAIT_AFTER_SEC_FLOAT):
    """Двойное нажатие левой клавиши мыши. Данное действие аналогично вызову функции (см. ниже).
    
    !ВНИМАНИЕ! Отсчет координат inXInt, inYInt начинается с левого верхнего края рабочей области (экрана).

    .. code-block:: python

        # Mouse: Взаимодействие с мышью
        from pyOpenRPA.Robot import Mouse
        Mouse.ClickDouble(100,150) #Выполнить двойное нажатие левой клавиши мыши на экране по координатам: X(гор) 100px, Y(вер) 150px.

    :param inXInt: Целевая позиция указателя мыши по оси X (горизонтальная ось). 
    :type inXInt: int, опциональный
    :param inYInt: Целевая позиция указателя мыши по оси Y (вертикальная ось). 
    :type inYInt: int, опциональный
    :param inButtonStr: Номер кнопки, которую требуется нажать. Возможные варианты: 'left', 'middle', 'right' или 1, 2, 3. В остальных случаях инициирует исключение ValueError. По умолчанию 'left'
    :type inButtonStr: str, опциональный
    :param inWaitAfterSecFloat: Количество секунд, которые ожидать после выполнения операции. По умолчанию установлено в настройках модуля Mouse (базовое значение 0.4)
    :type inWaitAfterSecFloat: float, опциональный
    """
    doubleClick(x=inXInt, y=inYInt)
    time.sleep(inWaitAfterSecFloat)

def Down(inXInt:int=None, inYInt:int=None, inButtonStr:str='left', inWaitAfterSecFloat:float=WAIT_AFTER_SEC_FLOAT):
    """Переместить указатель по координатам inXInt, inYInt, после чего нажать (вниз) клавишу мыши и не отпускать до выполнения соответсвующей команды (см. Up). Если координаты inXInt, inYInt не переданы - нажатие происходит на тех координатах X/Y, на которых указатель мыши находится.

    !ВНИМАНИЕ! Отсчет координат inXInt, inYInt начинается с левого верхнего края рабочей области (экрана).

    .. code-block:: python

        # Mouse: Взаимодействие с мышью
        from pyOpenRPA.Robot import Mouse
        Mouse.Down() #Опустить левую клавишу мыши

    :param inXInt: Целевая позиция указателя мыши по оси X (горизонтальная ось). 
    :type inXInt: int, опциональный
    :param inYInt: Целевая позиция указателя мыши по оси Y (вертикальная ось). 
    :type inYInt: int, опциональный
    :param inButtonStr: Номер кнопки, которую требуется нажать. Возможные варианты: 'left', 'middle', 'right' или 1, 2, 3. В остальных случаях инициирует исключение ValueError. По умолчанию 'left'
    :type inButtonStr: str, опциональный
    :param inWaitAfterSecFloat: Количество секунд, которые ожидать после выполнения операции. По умолчанию установлено в настройках модуля Mouse (базовое значение 0.4)
    :type inWaitAfterSecFloat: float, опциональный
    """
    mouseDown(x=inXInt, y=inYInt, button = inButtonStr)
    time.sleep(inWaitAfterSecFloat)

def Up(inXInt:int=None, inYInt:int=None, inButtonStr:str='left', inWaitAfterSecFloat:float=WAIT_AFTER_SEC_FLOAT):
    """Отпустить (вверх) клавишу мыши. Если координаты inXInt, inYInt не переданы - нажатие происходит на тех координатах X/Y, на которых указатель мыши находится.

    !ВНИМАНИЕ! Отсчет координат inXInt, inYInt начинается с левого верхнего края рабочей области.

    .. code-block:: python

        # Mouse: Взаимодействие с мышью
        from pyOpenRPA.Robot import Mouse
        Mouse.Up(inButtonStr:str='right') #Поднять правую клавишу мыши

    :param inXInt: Целевая позиция указателя мыши по оси X (горизонтальная ось). 
    :type inXInt: int, опциональный
    :param inYInt: Целевая позиция указателя мыши по оси Y (вертикальная ось). 
    :type inYInt: int, опциональный
    :param inButtonStr: Номер кнопки, которую требуется поднять. Возможные варианты: 'left', 'middle', 'right' или 1, 2, 3. В остальных случаях инициирует исключение ValueError. По умолчанию 'left'
    :type inButtonStr: str, опциональный
    :param inWaitAfterSecFloat: Количество секунд, которые ожидать после выполнения операции. По умолчанию установлено в настройках модуля Mouse (базовое значение 0.4)
    :type inWaitAfterSecFloat: float, опциональный
    """
    mouseUp(x=inXInt, y=inYInt, button = inButtonStr)
    time.sleep(inWaitAfterSecFloat)

def MoveTo(inXInt=None, inYInt=None, inMoveDurationSecFloat:float=0.0, inWaitAfterSecFloat:float=WAIT_AFTER_SEC_FLOAT):
    """Переместить указатель мыши на позицию inXInt, inYInt за время inMoveDurationSecFloat.

    !ВНИМАНИЕ! Отсчет координат inXInt, inYInt начинается с левого верхнего края рабочей области (экрана).

    .. code-block:: python

        # Mouse: Взаимодействие с мышью
        from pyOpenRPA.Robot import Mouse
        Mouse.MoveTo(inXInt=100, inYInt=200) #Переместить указатель мыши по координатам: X(гор) 100, Y(вер) 200

    :param inXInt: Целевая позиция указателя мыши по оси X (горизонтальная ось). 
    :type inXInt: int, опциональный
    :param inYInt: Целевая позиция указателя мыши по оси Y (вертикальная ось). 
    :type inYInt: int, опциональный
    :param inMoveDurationSecFloat: Время перемещения указателя мыши, По умолчанию 0.0 (моментальное перемещение)
    :type inMoveDurationSecFloat: float, опциональный
    :param inWaitAfterSecFloat: Количество секунд, которые ожидать после выполнения операции. По умолчанию установлено в настройках модуля Mouse (базовое значение 0.4)
    :type inWaitAfterSecFloat: float, опциональный
    """
    moveTo(x=inXInt, y=inYInt, duration=inMoveDurationSecFloat)
    time.sleep(inWaitAfterSecFloat)
    
def ScrollVertical(inScrollClickCountInt, inXInt=None, inYInt=None, inWaitAfterSecFloat:float=WAIT_AFTER_SEC_FLOAT):
    """Переместить указатель мыши на позицию inXInt, inYInt и выполнить вертикальную прокрутку (скроллинг) колесом мыши на количество щелчков inScrollClickCountInt.

    !ВНИМАНИЕ! Отсчет координат inXInt, inYInt начинается с левого верхнего края рабочей области (экрана).

    .. code-block:: python

        # Mouse: Взаимодействие с мышью
        from pyOpenRPA.Robot import Mouse
        Mouse.ScrollVertical(100, inXInt=100, inYInt=200) #Крутить колесо мыши вниз на 100 кликов по координатам: X(гор) 100, Y(вер) 200
        Mouse.ScrollVertical(-100) #Крутить колесо мыши вверх на 100 кликов по текущим координатам указателя мыши.

    :param inScrollClickCountInt: Количество щелчок колеса мыши, которое требуется !вертикально! прокрутить. Аргумент может принимать как положительное (прокрутка вниз), так и отрицательное (прокрутка вверх) значения
    :type inScrollClickCountInt: int, обязательный
    :param inXInt: Целевая позиция указателя мыши по оси X (горизонтальная ось). 
    :type inXInt: int, опциональный
    :param inYInt: Целевая позиция указателя мыши по оси Y (вертикальная ось). 
    :type inYInt: int, опциональный
    :param inWaitAfterSecFloat: Количество секунд, которые ожидать после выполнения операции. По умолчанию установлено в настройках модуля Mouse (базовое значение 0.4)
    :type inWaitAfterSecFloat: float, опциональный
    """
    vscroll(inScrollClickCountInt, x=inXInt, y=inYInt)
    time.sleep(inWaitAfterSecFloat)
    
def ScrollHorizontal(inScrollClickCountInt, inXInt=None, inYInt=None, inWaitAfterSecFloat:float=WAIT_AFTER_SEC_FLOAT):
    """!ТОЛЬКО ДЛЯ LINUX! Переместить указатель мыши на позицию inXInt, inYInt и выполнить горизонтальную прокрутку (скроллинг) виртуальным колесом мыши на количество щелчков inScrollClickCountInt.

    !ВНИМАНИЕ! Отсчет координат inXInt, inYInt начинается с левого верхнего края рабочей области (экрана).

    .. code-block:: python

        # Mouse: Взаимодействие с мышью
        from pyOpenRPA.Robot import Mouse
        Mouse.ScrollHorizontal(100, inXInt=100, inYInt=200) #Крутить колесо мыши вниз на 100 кликов по координатам: X(гор) 100, Y(вер) 200
        Mouse.ScrollHorizontal(-100) #Крутить колесо мыши вверх на 100 кликов по текущим координатам указателя мыши.

    :param inScrollClickCountInt: Количество щелчок колеса мыши, которое требуется !горизонтально! прокрутить. Аргумент может принимать как положительное (прокрутка вправо), так и отрицательное (прокрутка влево) значения
    :type inScrollClickCountInt: int, обязательный
    :param inXInt: Целевая позиция указателя мыши по оси X (горизонтальная ось). 
    :type inXInt: int, опциональный
    :param inYInt: Целевая позиция указателя мыши по оси Y (вертикальная ось). 
    :type inYInt: int, опциональный
    :param inWaitAfterSecFloat: Количество секунд, которые ожидать после выполнения операции. По умолчанию установлено в настройках модуля Mouse (базовое значение 0.4)
    :type inWaitAfterSecFloat: float, опциональный
    """
    hscroll(inScrollClickCountInt, x=inXInt, y=inYInt)
    time.sleep(inWaitAfterSecFloat)



# 1.2.0 - general processor - contains old orchestrator processor + RDPActive processor
import time, copy, threading
from . import A2O
# Run processor synchronious
def ProcessorRunSync(inGSettings):
    """
        "ProcessorDict": { # Has been changed. New general processor (one threaded) v.1.2.0
            "ActivityList": [ # List of the activities
                # {
                #    "Def":"DefAliasTest", # def link or def alias (look gSettings["Processor"]["AliasDefDict"])
                #    "ArgList":[1,2,3], # Args list
                #    "ArgDict":{"ttt":1,"222":2,"dsd":3}, # Args dictionary
                #    "ArgGSettings": None # Name of GSettings attribute: str (ArgDict) or index (for ArgList)
                #    "ArgLogger": None, # Name of GSettings attribute: str (ArgDict) or index (for ArgList)
                #    "GUIDStr": "sadasd-asdas-d-asdasd", # ActivityItem GUID which identify the Activity
                #    "ThreadBool": False
                # },
            ],
            "AliasDefDict": {}, # Storage for def with Str alias. To use it see pyOpenRPA.Orchestrator.ControlPanel
            "CheckIntervalSecFloat": 1.0 # Interval for check gSettings in ProcessorDict > ActivityList
            "ExecuteBool": True # Flag to execute thread processor
    """
    lL = inGSettings["Logger"]  # Logger alias
    inGSettings["ProcessorDict"]["ThreadIdInt"] = threading.get_ident() # fill Processor thread id
    while inGSettings["ProcessorDict"]["ExecuteBool"]:
        lActivityList = inGSettings["ProcessorDict"]["ActivityList"]  # Alias
        if len(lActivityList)>0:
            if lL: lL.debug(f'Processor ActivityList len: {len(lActivityList)}')
            lActivityItem = inGSettings["ProcessorDict"]["ActivityList"].pop(0) # Extract the first item from processor queue
            lResultList = ActivityListExecute(inGSettings = inGSettings, inActivityList = [lActivityItem]) # execute the activity item
            #Some help code
            if len(lResultList) == 0: lResultList= [None]
            # Send result to Orc if we have GUIDStr
            if "GUIDStr" in lActivityItem:
                # Def to send to Orc
                A2O.ActivityReturnDictSend(inGSettings=inGSettings, inActivityItemGUIDStr=lActivityItem["GUIDStr"],inReturn=lResultList[0])
        else:
            time.sleep(inGSettings["ProcessorDict"]["CheckIntervalSecFloat"]) # Sleep when list is empty

# Run processor Async
def ProcessorRunAsync(inGSettings, inActivityList):
    """
        "inActivityList": [ # List of the activities
            # {
            #    "Def":"DefAliasTest", # def link or def alias (look gSettings["Processor"]["AliasDefDict"])
            #    "ArgList":[1,2,3], # Args list
            #    "ArgDict":{"ttt":1,"222":2,"dsd":3}, # Args dictionary
            #    "ArgGSettings": None # Name of GSettings attribute: str (ArgDict) or index (for ArgList)
            #    "ArgLogger": None, # Name of GSettings attribute: str (ArgDict) or index (for ArgList)
            #    "GUIDStr": "sadasd-asdas-d-asdasd", # ActivityItem GUID which identify the Activity
            #    "ThreadBool": True
            # },
    """
    def __process__(inGSettings, inActivityList):
        for lActivityItem in inActivityList:
            lL = inGSettings["Logger"]  # Logger alias
            if lL: lL.debug(f'ActivityItem in new thread')
            lResultList = ActivityListExecute(inGSettings = inGSettings, inActivityList = [lActivityItem]) # execute the activity item
            #Some help code
            if len(lResultList) == 0: lResultList= [None]
            # Send result to Orc if we have GUIDStr
            if "GUIDStr" in lActivityItem:
                # Def to send to Orc
                A2O.ActivityReturnDictSend(inGSettings=inGSettings, inActivityItemGUIDStr=lActivityItem["GUIDStr"],inReturn=lResultList[0])
    # Start in new thread
    lThread = threading.Thread(target=__process__,kwargs={"inGSettings": inGSettings, "inActivityList": inActivityList})
    lThread.start()

# Execute ActivityItem list
# return the def result
def ActivityListExecute(inGSettings, inActivityList):
    lL = inGSettings["Logger"]  # Logger alias
    lResultList = [] # init the result list
    for lActivityItem in inActivityList:  # Iterate throught the activity list
        lDef = None  # Def variable
        if callable(lActivityItem["Def"]):  # CHeck if def is callable
            lDef = lActivityItem["Def"]  # Get the def
        else:  # Is not callable - check alias
            lDef = inGSettings["ProcessorDict"]["AliasDefDict"].get(lActivityItem["Def"], None)  # get def if def key in Alias def storage
        #gSettings
        lGSettingsDictKey = lActivityItem.pop("ArgGSettings",None)
        # # Prepare arg dict - gSettings
        if type(lGSettingsDictKey) is str: # check if gSetting key is in ArgDict
            lActivityItem["ArgDict"][lGSettingsDictKey] = inGSettings # Set the gSettings in dict
        # # Prepare arg list
        elif type(lGSettingsDictKey) is int: # check if gSetting key is in ArgDict
            lActivityItem["ArgList"].insert(lGSettingsDictKey,inGSettings)# Set the gSettings in list by the index
        #Logger
        lLoggerDictKey = lActivityItem.pop("ArgLogger",None)
        # # Prepare arg dict - gSettings
        if type(lLoggerDictKey) is str: # check if gSetting key is in ArgDict
            lActivityItem["ArgDict"][lLoggerDictKey] = lL # Set the lLogger in dict
        # # Prepare arg list
        elif type(lLoggerDictKey) is int: # check if gSetting key is in ArgDict
            lActivityItem["ArgList"].insert(lLoggerDictKey,lL)# Set the lLogger in list by the index

        try:  # try to run function from Processor.py
            lActivityItemResult = lDef(*lActivityItem["ArgList"], **lActivityItem["ArgDict"])
            lResultList.append(lActivityItemResult) # return the result
        except Exception as e:
            if lL: lL.exception(f"Processor.ActivityListExecute: Exception in def execution - activity will be ignored. Activity item: {lActivityItem}")  # Logging
            lResultList.append(None) # return the generated exception
    return lResultList # return the result list


import time
import os
from .. import __Orchestrator__
from . import Process
import threading
from typing import List
from typing import Tuple

from pyOpenRPA import Orchestrator

class Git():

    mAgentHostNameStr = None
    mAgentUserNameStr = None
    mAbsPathStr = None
    mProcessList: List[Tuple] = [] # List of the key turples of the Process instance

    def __init__(self, inAgentHostNameStr=None, inAgentUserNameStr=None, inGitPathStr=""):
        """
        Init the Git repo instance. It helps to detect new changes in repo and auto restart services

        :param inAgentHostNameStr: Agent hostname in any case. Required to identify Process. If None - works with Orc session
        :param inAgentUserNameStr: Agent user name in any case. Required to identify Process. If None - works with Orc session
        :param inGitPathStr: Relative (from the orchestrator working directory) or absolute. If "" - work with Orc repo
        :return:
        """
        lAbsPathStr = os.path.abspath(inGitPathStr)
        lAbsPathUpperStr = lAbsPathStr.upper()
        lGS = __Orchestrator__.GSettingsGet()
        # Check if Process is not exists in GSettings
        if (inAgentHostNameStr.upper(), inAgentUserNameStr.upper(), lAbsPathUpperStr) not in lGS["ManagersGitDict"]:
            self.mAbsPathStr = lAbsPathStr
            self.mAbsPathUpperStr = lAbsPathUpperStr
            self.mAgentHostNameStr = inAgentHostNameStr
            self.mAgentUserNameStr = inAgentUserNameStr
            lGS["ManagersGitDict"][(inAgentHostNameStr.upper(), inAgentUserNameStr.upper(), lAbsPathUpperStr)]=self
        else: raise Exception(f"Managers.Git ({inAgentHostNameStr}, {inAgentUserNameStr}, {lAbsPathUpperStr}): Can't init the Git instance because it already inited in early")

    def ProcessConnect(self, inProcess: Process):
        """
        Connect process to the Git instance. It will apply to stop safe process when upgrade the repo and than start it

        :param inProcess: Process instance
        :type inProcess: Process
        """
        lProcessTurple = inProcess.KeyTurpleGet()
        if lProcessTurple not in self.mProcessList:
            self.mProcessList.append(lProcessTurple)
        else:
            raise Exception(f"Process with current key is already exists in Git process list.")

    def ProcessListSaveStopSafe(self):
        """
        Save the state and do the stop safe for the all processes
        Will send safe stop in parallel mode but wait to the end of the safestop for the all processes. After that will continue
        """
        lIntervalScheckSecFloat = 5.0
        lThreadList:List[threading.Thread] = []
        for lProcessItemTuple in self.mProcessList:
            lProcessItem = Orchestrator.Managers.ProcessGet(*lProcessItemTuple)
            lProcessItem.StatusSave()
            lThread = threading.Thread(target=lProcessItem.StopSafe)
            lThread.start()
            lThreadList.append(lThread)
        # Wait for all process will be safe stopped
        lAllThreadStoppedBool = False
        while not lAllThreadStoppedBool:
            lAllThreadStoppedBool = True
            for lThread in lThreadList:
                if lThread.is_alive() == True:
                    lAllThreadStoppedBool = False
                    break
            time.sleep(lIntervalScheckSecFloat)

    def ProcessListRestore(self):
        """
        Restore the process state for the all processes
        """
        for lProcessItem in self.mProcessList:
            lProcessItem.StatusRestore()

    def __OSCMDShell__(self, inCMDStr):
        """
        Detect the way of use and send the cmd. Wait for command execution!

        :return: None is not exists
        """
        if self.mAgentUserNameStr is not None and self.mAgentHostNameStr is not None: # Check if Agent specified
            lActivityItemGUIDStr = __Orchestrator__.AgentOSCMD(inHostNameStr=self.mAgentHostNameStr,inUserStr=self.mAgentUserNameStr,inCMDStr=inCMDStr,inRunAsyncBool=False,inSendOutputToOrchestratorLogsBool=False)
            lCMDResultStr = __Orchestrator__.AgentActivityItemReturnGet(inGUIDStr=lActivityItemGUIDStr)
        else:
            lCMDResultStr = __Orchestrator__.OSCMD(inCMDStr=inCMDStr, inRunAsyncBool=False)
        return lCMDResultStr

    def BranchRevGet(self, inBranchNameStr="HEAD"):
        """
        Get the specified branch revision. Default return the current branch revision

        .. code-block:: python
            lGit.BranchRevGet(inBranchNameStr="dev") # Get revision of the local dev branch
            lGit.BranchRevGet(inBranchNameStr="remotes/origin/dev") # Get revision of the remotes dev branch
            lGit.BranchRevGet(inBranchNameStr="HEAD") # Get revision of the current HEAD branch
            lGit.BranchRevGet() # Equal to the call inBranchNameStr="HEAD"

        :param inBranchNameStr: The branch name where to get revision guid
        :return: revision GUID
        """
        lCMDStr = f"cd \"{self.mAbsPathUpperStr}\" && git rev-parse {inBranchNameStr}"
        lCMDResultStr = self.__OSCMDShell__(inCMDStr=lCMDStr)
        return lCMDResultStr

    def BranchRevIsLast(self, inBranchLocalStr: str, inBranchRemoteStr: str) -> bool:
        """Get fetch and check if local branch revision is last (if check with remote)

        :param inBranchLocalStr: _description_
        :type inBranchLocalStr: str
        :param inBranchRemoteStr: example: origin/prd
        :type inBranchRemoteStr: str
        :return: _description_
        :rtype: bool
        """
        lIsLastBool = False
        self.Fetch()
        lLocalBranchRevStr = self.BranchRevGet(inBranchNameStr=inBranchLocalStr)
        lRemoteBranchRevStr = self.BranchRevGet(inBranchNameStr=inBranchRemoteStr)
        if lLocalBranchRevStr == lRemoteBranchRevStr:
            lIsLastBool = True
        return lIsLastBool


    def BranchRevLastGetInterval(self, inBranchLocalStr: str, inBranchRemoteStr: str, inPreviousBranchRestoreBool: bool = True, inIntervalSecFloat: float = 60.0):
        """Periodically check if revision is last

        :param inBranchLocalStr: _description_
        :type inBranchLocalStr: str
        :param inBranchRemoteStr: example: origin/prd
        :type inBranchRemoteStr: str
        :param inPreviousBranchRestoreBool: _description_, defaults to True
        :type inPreviousBranchRestoreBool: bool, optional
        :param inIntervalSecFloat: _description_, defaults to 60.0
        :type inIntervalSecFloat: float, optional
        """    
        #self.BranchRevLastGet(inBranchLocalStr, inBranchRemoteStr, inPreviousBranchRestoreBool)    
        Orchestrator.OrchestratorScheduleGet().every(inIntervalSecFloat).seconds.do(self.BranchRevLastGet, inBranchLocalStr, inBranchRemoteStr, inPreviousBranchRestoreBool)

    def BranchRevLastGet(self, inBranchLocalStr: str, inBranchRemoteStr: str, inPreviousBranchRestoreBool: bool = True):
        """Do some action to get the last revision

        :param inBranchLocalStr: [description]
        :type inBranchLocalStr: str
        :param inBranchRemoteStr: [description]
        :type inBranchRemoteStr: str
        """
        Orchestrator.OrchestratorLoggerGet().debug(f"Managers.Git ({self.mAbsPathStr}): self.BranchRevLastGet has been init")
        # check if the correct revision
        lCMDResultStr = None
        if self.BranchRevIsLast(inBranchLocalStr=inBranchLocalStr, inBranchRemoteStr=inBranchRemoteStr) == False:
            Orchestrator.OrchestratorLoggerGet().info(f"Managers.Git ({self.mAbsPathStr}): self.BranchRevLastGet, new rev (branch: {inBranchLocalStr}) has been detected - merge (branch: {inBranchRemoteStr})")
            # Do the stop safe for the connected process
            self.ProcessListSaveStopSafe()
            lBranchNameCurrentStr = self.BranchNameGet()
            # reset all changes in local folder
            self.Clear()
            # checkout
            self.BranchCheckout(inBranchNameStr=inBranchLocalStr)
            # merge
            lCMDStr = f"cd \"{self.mAbsPathUpperStr}\" && git merge {inBranchRemoteStr}"
            lCMDResultStr = self.__OSCMDShell__(inCMDStr=lCMDStr)
            if inPreviousBranchRestoreBool == True:
                # checkout to the source branch which was
                self.BranchCheckout(inBranchNameStr=lBranchNameCurrentStr)
            # do the orc restart
            Orchestrator.OrchestratorLoggerGet().info(f"Managers.Git ({self.mAbsPathStr}): self.BranchRevLastGet, merge done, restart orc")
            Orchestrator.OrchestratorRestart()
        return lCMDResultStr

    def BranchNameGet(self) -> str:
        """Get the current local branch name

        :return: current local branch name
        """
        #"git rev-parse --abbrev-ref HEAD"
        lCMDStr = f"cd \"{self.mAbsPathUpperStr}\" && git rev-parse --abbrev-ref HEAD"
        lCMDResultStr = self.__OSCMDShell__(inCMDStr=lCMDStr)
        return lCMDResultStr

    def BranchCheckout(self, inBranchNameStr):
        self.Clear()
        lCMDStr = f"cd \"{self.mAbsPathUpperStr}\" && git checkout {inBranchNameStr}"
        lCMDResultStr = self.__OSCMDShell__(inCMDStr=lCMDStr)
        return lCMDResultStr

    def Clear(self):
        """Clear the all changes in the local folder. Get up to the current revision
        """
        # f"git clean -f -d" # Очистить от лишних файлов
        lCMDStr = f"cd \"{self.mAbsPathUpperStr}\" && git clean -f -d"
        lCMDResultStr = self.__OSCMDShell__(inCMDStr=lCMDStr)
        # f"git reset --hard" # Откатить файлы, которые отслеживаются Git и которые были изменены
        lCMDStr = f"cd \"{self.mAbsPathUpperStr}\" && git reset --hard"
        lCMDResultStr = self.__OSCMDShell__(inCMDStr=lCMDStr)
        return lCMDResultStr

    def Fetch(self):
        """
        Get updates from the git server.

        .. code-block:: python
            lGit.Fetch() # get updates from the server

        :return: None
        """
        lCMDStr = f"cd \"{self.mAbsPathUpperStr}\" && git fetch"
        lCMDResultStr = self.__OSCMDShell__(inCMDStr=lCMDStr)
        return lCMDResultStr

def GitExists(inAgentHostNameStr: str, inAgentUserNameStr: str, inGitPathStr: str) -> bool:
    """
    Check if the Git instance is exists in GSettings by the (inAgentHostNameStr: str, inAgentUserNameStr: str, inGitPathStr: str)

    :param inAgentHostNameStr: Agent hostname in any case. Required to identify Process
    :param inAgentUserNameStr: Agent user name in any case. Required to identify Process
    :param inGitPathStr: Relative (from the orchestrator working directory) or absolute. If "" - work with Orc repo
    :return: True - process exists in gsettings; False - else
    """
    return (inAgentHostNameStr.upper(), inAgentUserNameStr.upper(), inGitPathStr.upper()) in __Orchestrator__.GSettingsGet()["ManagersGitDict"]


def GitGet(inAgentHostNameStr: str, inAgentUserNameStr: str, inGitPathStr: str) -> Git:
    """
    Return the Git instance by the (inAgentHostNameStr: str, inAgentUserNameStr: str, inGitPathStr: str)

    :param inAgentHostNameStr: Agent hostname in any case. Required to identify Process
    :param inAgentUserNameStr: Agent user name in any case. Required to identify Process
    :param inGitPathStr: Relative (from the orchestrator working directory) or absolute. If "" - work with Orc repo
    :return: Git instance (if exists) Else None
    """
    lAbsPathUpperStr = os.path.abspath(inGitPathStr).upper()
    return __Orchestrator__.GSettingsGet()["ManagersGitDict"].get((inAgentHostNameStr.upper(), inAgentUserNameStr.upper(), lAbsPathUpperStr),None)


def GitBranchRevGet(inAgentHostNameStr: str, inAgentUserNameStr: str, inGitPathStr: str, inBranchNameStr: str="HEAD") -> str:
    lGit = GitGet(inAgentHostNameStr = inAgentHostNameStr, inAgentUserNameStr = inAgentUserNameStr, inGitPathStr=inGitPathStr)
    if lGit is not None: return lGit.BranchRevGet(inBranchNameStr=inBranchNameStr)

def GitFetch(inAgentHostNameStr: str, inAgentUserNameStr: str, inGitPathStr: str) -> None:
    lGit = GitGet(inAgentHostNameStr=inAgentHostNameStr, inAgentUserNameStr=inAgentUserNameStr,
                  inGitPathStr=inGitPathStr)
    if lGit is not None: lGit.Fetch()


# Def to check inGSettings and update structure to the backward compatibility
# !!! ATTENTION: Backward compatibility has been started from v1.1.13 !!!
# So you can use config of the orchestrator 1.1.13 in new Orchestrator versions and all will be ok :) (hope it's true)
import win32security, json, datetime, time, copy
import schedule
# # # # # # # # # # # # # # # # # # #
# Backward compatibility Web defs up to v1.2.0
# # # # # # # # # # # # # # # # # # #

# UserAccess get rights hierarchy dict in json
def v1_2_0_UserRoleHierarchyGet(inRequest,inGlobalDict):
    inResponseDict = inRequest.OpenRPAResponseDict
    # Create result JSON
    lResultDict = inRequest.OpenRPA["DefUserRoleHierarchyGet"]() # Get the User Role Hierarchy list
    # Send message back to client
    message = json.dumps(lResultDict)
    # Write content as utf-8 data
    inResponseDict["Body"] = bytes(message, "utf8")

from inspect import signature # For detect count of def args
# /Orchestrator/RobotRDPActive/ControlPanelDictGet
def v1_2_0_RobotRDPActive_ControlPanelDictGet(inRequest,inGlobalDict):
    inResponseDict = inRequest.OpenRPAResponseDict
    lResultDict = {
        "DataList":[
            # {"SessionKeyStr":"", "SessionHexStr": "", "IsFullScreenBool": False, "IsIgnoredBool": False}
        ]
    }
    # Iterate throught the RDP List
    for lRDPSessionKeyStrItem in inGlobalDict["RobotRDPActive"]["RDPList"]:
        lRDPConfiguration = inGlobalDict["RobotRDPActive"]["RDPList"][lRDPSessionKeyStrItem] # Get the configuration dict
        lDataItemDict = {"SessionKeyStr":"", "SessionHexStr": "", "IsFullScreenBool": False, "IsIgnoredBool": False} # Template
        lDataItemDict["SessionKeyStr"] = lRDPSessionKeyStrItem # Session key str
        lDataItemDict["SessionHexStr"] = lRDPConfiguration["SessionHex"] # Session Hex
        lDataItemDict["IsFullScreenBool"] = True if lRDPSessionKeyStrItem == inGlobalDict["RobotRDPActive"]["FullScreenRDPSessionKeyStr"] else False # Check  the full screen for rdp window
        lDataItemDict["IsIgnoredBool"] = lRDPConfiguration["SessionIsIgnoredBool"] # Is ignored
        lResultDict["DataList"].append(lDataItemDict)
    # Send message back to client
    message = json.dumps(lResultDict)
    # Write content as utf-8 data
    inResponseDict["Body"] = bytes(message, "utf8")

# def to check control panels for selected session
def v1_2_0_Monitor_ControlPanelDictGet_SessionCheckInit(inRequest,inGlobalDict):
    lL = inGlobalDict["Logger"]  # Alias for logger
    lLifetimeSecFloat = inGlobalDict["Client"]["Session"]["LifetimeSecFloat"]
    lLifetimeRequestSecFloat = inGlobalDict["Client"]["Session"]["LifetimeRequestSecFloat"]
    lControlPanelRefreshIntervalSecFloat = inGlobalDict["Client"]["Session"]["ControlPanelRefreshIntervalSecFloat"]
    lCookieSessionGUIDStr = None  # generate the new GUID
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    # Technicaldef - interval check control panels + check actuality of the session by the datetime
    def TechnicalCheck():
        lItemValue = inGlobalDict["Client"]["Session"]["TechnicalSessionGUIDCache"][lCookieSessionGUIDStr]
        # Lifetime is ok - check control panel
        lDatasetCurrentBytes = v1_2_0_Monitor_ControlPanelDictGet(inRequest,inGlobalDict) # Call the control panel
        if lDatasetCurrentBytes != lItemValue["DatasetLast"]["ControlPanel"]["Data"]: # Check if dataset is changed
            lItemValue["DatasetLast"]["ControlPanel"]["Data"] = lDatasetCurrentBytes # Set new datset
            lItemValue["DatasetLast"]["ControlPanel"]["ReturnBool"] = True # Set flag to return the data
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    # Technicaldef - Create new session struct
    def TechnicalSessionNew(inSessionGUIDStr):
        lCookieSessionGUIDStr = inSessionGUIDStr  # Generate the new GUID
        lSessionNew = { # Session with some GUID str. On client session guid stored in cookie "SessionGUIDStr"
            "InitDatetime": datetime.datetime.now(), # Datetime when session GUID was created
            "DatasetLast": {
                "ControlPanel": {
                    "Data": None, # Struct to check with new iterations. None if starts
                    "ReturnBool": False # flag to return, close request and return data as json
                }
            },
            "ClientRequestHandler": inRequest, # Last client request handler
            "UserADStr": inRequest.OpenRPA["User"], # User, who connect. None if user is not exists
            "DomainADStr": inRequest.OpenRPA["Domain"], # Domain of the user who connect. None if user is not exists
        }
        inGlobalDict["Client"]["Session"]["TechnicalSessionGUIDCache"][lCookieSessionGUIDStr] = lSessionNew # Set new session in dict
        inRequest.OpenRPAResponseDict["SetCookies"]["SessionGUIDStr"] = lCookieSessionGUIDStr # Set SessionGUIDStr in cookies
        if lL: lL.info(f"New session GUID is created. GUID {lCookieSessionGUIDStr}")
        return lCookieSessionGUIDStr
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    lCreateNewSessionBool = False # Flag to create new session structure
    # step 1 - get cookie SessionGUIDStr
    lSessionGUIDStr = inRequest.headers.get("SessionGUIDStr", None)
    if lSessionGUIDStr is not None: # Check if GUID session is ok
        #inRequest.OpenRPAResponseDict["StatusCode"] = 301
        #inRequest.OpenRPAResponseDict["Headers"]["Location"] = "/"
        #if lL: lL.info(f"GUID is detected - send HTTP 301 to refresh page")
        lCookieSessionGUIDStr = lSessionGUIDStr # Get the existing GUID
        if lSessionGUIDStr not in inGlobalDict["Client"]["Session"]["TechnicalSessionGUIDCache"]:
            lCookieSessionGUIDStr= TechnicalSessionNew(inSessionGUIDStr = lSessionGUIDStr) # Create new session
        else: # Update the datetime of the request session
            inGlobalDict["Client"]["Session"]["TechnicalSessionGUIDCache"][lCookieSessionGUIDStr]["InitDatetime"]=datetime.datetime.now()
    else:
        lCookieSessionGUIDStr = TechnicalSessionNew(inSessionGUIDStr = lSessionGUIDStr) # Create new session
    # Init the RobotRDPActive in another thread
    #lThreadCheckCPInterval = threading.Thread(target=TechnicalIntervalCheck)
    #lThreadCheckCPInterval.daemon = True  # Run the thread in daemon mode.
    #lThreadCheckCPInterval.start()  # Start the thread execution.

    # Step 2 - interval check if data is exist
    lTimeStartSecFloat = time.time()
    lDoWhileBool = True # Flag to iterate throught the lifetime of the request
    while lDoWhileBool:
        #print(lTechnicalSessionGUIDCache)
        #print(lCookieSessionGUIDStr)
        if lCookieSessionGUIDStr in inGlobalDict["Client"]["Session"]["TechnicalSessionGUIDCache"]:
            lItemValue = inGlobalDict["Client"]["Session"]["TechnicalSessionGUIDCache"][lCookieSessionGUIDStr]
            if (time.time() - lTimeStartSecFloat) >= lLifetimeRequestSecFloat: # Check if lifetime client request is over or has no key
                if lL: lL.debug(f"Client request lifetime is over")
                lDoWhileBool = False # Stop the iterations
            if lDoWhileBool:
                TechnicalCheck() # Calculate the CP
                if lItemValue["DatasetLast"]["ControlPanel"]["ReturnBool"] == True: # Return data if data flag it True
                    lDatasetCurrentBytes = lItemValue["DatasetLast"]["ControlPanel"]["Data"]  # Set new dataset
                    inResponseDict = inRequest.OpenRPAResponseDict
                    inResponseDict["Body"] = lDatasetCurrentBytes
                    lItemValue["DatasetLast"]["ControlPanel"]["ReturnBool"] = False  # Set flag that data was returned
                    lDoWhileBool = False  # Stop the iterations
        else:
            lCookieSessionGUIDStr = TechnicalSessionNew(inSessionGUIDStr = lCookieSessionGUIDStr) # Create new session
        if lDoWhileBool:  # Sleep if we wait hte next iteration
            time.sleep(lControlPanelRefreshIntervalSecFloat)  # Sleep to the next iteration

def v1_2_0_Monitor_ControlPanelDictGet(inRequest,inGlobalDict):
    inResponseDict = inRequest.OpenRPAResponseDict
    lL = inGlobalDict["Logger"] # Alias for logger
    # Create result JSON
    lResultJSON = {"RenderRobotList": [], "RenderRDPList": []}
    lRenderFunctionsRobotList = inGlobalDict["ControlPanelDict"]["RobotList"]
    for lItem in lRenderFunctionsRobotList:
        lUACBool = True # Check if render function is applicable User Access Rights (UAC)
        if inGlobalDict["Server"]["AccessUsers"]["FlagCredentialsAsk"] is True:
            lUserRights = inGlobalDict["Server"]["AccessUsers"]["RuleDomainUserDict"][(inRequest.OpenRPA["Domain"].upper(),inRequest.OpenRPA["User"].upper())]
            if len(lUserRights["ControlPanelKeyAllowedList"]) > 0 and lItem["KeyStr"] not in lUserRights["ControlPanelKeyAllowedList"]:
                lUACBool = False # UAC Check is not passed - False for user
        if lUACBool: # Run function if UAC is TRUE
            # Выполнить вызов и записать результат
            # Call def (inRequest, inGSettings) or def (inGSettings)
            lItemResultDict = None
            lDEFSignature = signature(lItem["RenderFunction"]) # Get signature of the def
            lDEFARGLen = len(lDEFSignature.parameters.keys()) # get count of the def args
            try:
                if lDEFARGLen == 1: # def (inGSettings)
                    lItemResultDict = lItem["RenderFunction"](inGlobalDict)
                elif lDEFARGLen == 2: # def (inRequest, inGSettings)
                    lItemResultDict = lItem["RenderFunction"](inRequest, inGlobalDict)
                elif lDEFARGLen == 0: # def ()
                    lItemResultDict = lItem["RenderFunction"]()
                # RunFunction
                lResultJSON["RenderRobotList"].append(lItemResultDict)
            except Exception as e:
                if lL: lL.exception(f"Error in control panel. CP item {lItem}. Exception is below")
    # Iterate throught the RDP list
    for lRDPSessionKeyStrItem in inGlobalDict["RobotRDPActive"]["RDPList"]:
        lRDPConfiguration = inGlobalDict["RobotRDPActive"]["RDPList"][
            lRDPSessionKeyStrItem]  # Get the configuration dict
        lDataItemDict = {"SessionKeyStr": "", "SessionHexStr": "", "IsFullScreenBool": False,
                         "IsIgnoredBool": False}  # Template
        lDataItemDict["SessionKeyStr"] = lRDPSessionKeyStrItem  # Session key str
        lDataItemDict["SessionHexStr"] = lRDPConfiguration["SessionHex"]  # Session Hex
        lDataItemDict["IsFullScreenBool"] = True if lRDPSessionKeyStrItem == inGlobalDict["RobotRDPActive"][
            "FullScreenRDPSessionKeyStr"] else False  # Check  the full screen for rdp window
        lDataItemDict["IsIgnoredBool"] = lRDPConfiguration["SessionIsIgnoredBool"]  # Is ignored
        lResultJSON["RenderRDPList"].append(lDataItemDict)
    # Send message back to client
    message = json.dumps(lResultJSON)
    # Write content as utf-8 data
    #inResponseDict["Body"] = bytes(message, "utf8")
    return bytes(message, "utf8")


from . import __Orchestrator__ # For user defs



# v1.2.0 Def for old procesor to new processor
# Return new activity for the new processor
def v1_2_0_ProcessorOld2NewActivityDict(inActivityOld):
    if inActivityOld["Type"] == "WindowsLogon":
        lResult = {
            "Def": __Orchestrator__.OSCredentialsVerify, # def link or def alias (look gSettings["Processor"]["AliasDefDict"])
            "ArgList":[], # Args list
            "ArgDict":{"inUserStr": inActivityOld["User"],"inPasswordStr":inActivityOld["Password"],"inDomainStr":inActivityOld["Domain"]}, # Args dictionary
            "ArgGSettings": None,  # Name of GSettings attribute: str (ArgDict) or index (for ArgList)
            "ArgLogger": None  # Name of GSettings attribute: str (ArgDict) or index (for ArgList)
        }
    elif inActivityOld["Type"] == "GlobalDictKeyListValueGet":
        lResult = {
            "Def": __Orchestrator__.GSettingsKeyListValueGet, # def link or def alias (look gSettings["Processor"]["AliasDefDict"])
            "ArgList":[], # Args list
            "ArgDict":{"inKeyList": inActivityOld["KeyList"]}, # Args dictionary
            "ArgGSettings": "inGSettings",  # Name of GSettings attribute: str (ArgDict) or index (for ArgList)
            "ArgLogger": None  # Name of GSettings attribute: str (ArgDict) or index (for ArgList)
        }
    elif inActivityOld["Type"] == "CMDStart":
        lResult = {
            "Def": __Orchestrator__.OSCMD, # def link or def alias (look gSettings["Processor"]["AliasDefDict"])
            "ArgList": [],  # Args list
            "ArgDict": {"inCMDStr": inActivityOld["Command"]},  # Args dictionary
            "ArgGSettings": None,  # Name of GSettings attribute: str (ArgDict) or index (for ArgList)
            "ArgLogger": "inLogger"  # Name of GSettings attribute: str (ArgDict) or index (for ArgList)
        }
    elif inActivityOld["Type"] == "OrchestratorRestart":
        lResult = {
            "Def": __Orchestrator__.OrchestratorRestart, # def link or def alias (look gSettings["Processor"]["AliasDefDict"])
            "ArgList": [],  # Args list
            "ArgDict": {},  # Args dictionary
            "ArgGSettings": "inGSettings",  # Name of GSettings attribute: str (ArgDict) or index (for ArgList)
            "ArgLogger": None  # Name of GSettings attribute: str (ArgDict) or index (for ArgList)
        }
    elif inActivityOld["Type"] == "OrchestratorSessionSave":
        lResult = {
            "Def": __Orchestrator__.OrchestratorSessionSave,
            # def link or def alias (look gSettings["Processor"]["AliasDefDict"])
            "ArgList": [],  # Args list
            "ArgDict": {},  # Args dictionary
            "ArgGSettings": "inGSettings",  # Name of GSettings attribute: str (ArgDict) or index (for ArgList)
            "ArgLogger": None  # Name of GSettings attribute: str (ArgDict) or index (for ArgList)
        }
    elif inActivityOld["Type"] == "GlobalDictKeyListValueSet":
        lResult = {
            "Def": __Orchestrator__.GSettingsKeyListValueSet, # def link or def alias (look gSettings["Processor"]["AliasDefDict"])
            "ArgList": [],  # Args list
            "ArgDict": {"inKeyList": inActivityOld["KeyList"], "inValue": inActivityOld["Value"]},  # Args dictionary
            "ArgGSettings": "inGSettings",  # Name of GSettings attribute: str (ArgDict) or index (for ArgList)
            "ArgLogger": None  # Name of GSettings attribute: str (ArgDict) or index (for ArgList)
        }
    elif inActivityOld["Type"] == "GlobalDictKeyListValueAppend":
        lResult = {
            "Def": __Orchestrator__.GSettingsKeyListValueAppend, # def link or def alias (look gSettings["Processor"]["AliasDefDict"])
            "ArgList": [],  # Args list
            "ArgDict": {"inKeyList": inActivityOld["KeyList"], "inValue": inActivityOld["Value"]},  # Args dictionary
            "ArgGSettings": "inGSettings",  # Name of GSettings attribute: str (ArgDict) or index (for ArgList)
            "ArgLogger": None  # Name of GSettings attribute: str (ArgDict) or index (for ArgList)
        }
    elif inActivityOld["Type"] == "GlobalDictKeyListValueOperator+":
        lResult = {
            "Def": __Orchestrator__.GSettingsKeyListValueOperatorPlus, # def link or def alias (look gSettings["Processor"]["AliasDefDict"])
            "ArgList": [],  # Args list
            "ArgDict": {"inKeyList": inActivityOld["KeyList"], "inValue": inActivityOld["Value"]},  # Args dictionary
            "ArgGSettings": "inGSettings",  # Name of GSettings attribute: str (ArgDict) or index (for ArgList)
            "ArgLogger": None  # Name of GSettings attribute: str (ArgDict) or index (for ArgList)
        }
    elif inActivityOld["Type"] == "ProcessStart":
        lResult = {
            "Def": __Orchestrator__.ProcessStart, # def link or def alias (look gSettings["Processor"]["AliasDefDict"])
            "ArgList": [],  # Args list
            "ArgDict": {"inPathStr": inActivityOld["Path"], "inArgList": inActivityOld["ArgList"]},  # Args dictionary
            "ArgGSettings": None,  # Name of GSettings attribute: str (ArgDict) or index (for ArgList)
            "ArgLogger": None  # Name of GSettings attribute: str (ArgDict) or index (for ArgList)
        }
    elif inActivityOld["Type"] == "ProcessStartIfTurnedOff":
        lResult = {
            "Def": __Orchestrator__.ProcessStart, # def link or def alias (look gSettings["Processor"]["AliasDefDict"])
            "ArgList": [],  # Args list
            "ArgDict": {"inPathStr": inActivityOld["Path"], "inArgList": inActivityOld["ArgList"], "inStopProcessNameWOExeStr": inActivityOld["CheckTaskName"]},  # Args dictionary
            "ArgGSettings": None,  # Name of GSettings attribute: str (ArgDict) or index (for ArgList)
            "ArgLogger": None  # Name of GSettings attribute: str (ArgDict) or index (for ArgList)
        }
    elif inActivityOld["Type"] == "ProcessStop":
        lResult = {
            "Def": __Orchestrator__.ProcessStop, # def link or def alias (look gSettings["Processor"]["AliasDefDict"])
            "ArgList": [],  # Args list
            "ArgDict": {"inProcessNameWOExeStr": inActivityOld["Name"], "inCloseForceBool": inActivityOld["FlagForce"], "inUserNameStr": inActivityOld["User"]},  # Args dictionary
            "ArgGSettings": None,  # Name of GSettings attribute: str (ArgDict) or index (for ArgList)
            "ArgLogger": None  # Name of GSettings attribute: str (ArgDict) or index (for ArgList)
        }
    elif inActivityOld["Type"] == "PythonStart":
        lResult = {
            "Def": __Orchestrator__.PythonStart, # def link or def alias (look gSettings["Processor"]["AliasDefDict"])
            "ArgList": [],  # Args list
            "ArgDict": {"inModulePathStr": inActivityOld["ModuleName"], "inDefNameStr": inActivityOld["FunctionName"], "inArgList": inActivityOld["ArgList"],
                    "inArgDict": inActivityOld["ArgDict"] },  # Args dictionary
            "ArgGSettings": None,  # Name of GSettings attribute: str (ArgDict) or index (for ArgList)
            "ArgLogger": "inLogger"  # Name of GSettings attribute: str (ArgDict) or index (for ArgList)
        }
    else:
        raise Exception(f"BackwardCompatibility up to v1.2.0, old processor: No type {inActivityOld['Type']} has been found in old processor.")
    return lResult # return the result


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# # # # # # # # # HERE IS THE MAIN DEF WHICH IS LAUNCHES WHEN START # # # # # # # # #
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
def Update(inGSettings):
    lL = inGSettings["Logger"] # Alias for logger
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    # v1.1.13 to v1.1.14
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    if "Autocleaner" not in inGSettings: # Add "Autocleaner" structure
        inGSettings["Autocleaner"] = { # Some gurbage is collecting in g settings. So you can configure autocleaner to periodically clear gSettings
            "IntervalSecFloat": 7200.0, # Sec float to periodically clear gsettings
        }
        if lL: lL.warning(f"Backward compatibility (v1.1.13 to v1.1.14): Add default 'Autocleaner' structure") # Log about compatibility
    if "Client" not in inGSettings: # Add "Client" structure
        inGSettings["Client"] = { # Settings about client web orchestrator
            "Session":{ # Settings about web session. Session algorythms works only for special requests (URL in ServerSettings)
                "LifetimeSecFloat": 600.0, # Client Session lifetime in seconds. after this time server will forget about this client session
                "LifetimeRequestSecFloat": 120.0, # 1 client request lifetime in server in seconds
                "ControlPanelRefreshIntervalSecFloat": 1.5, # Interval to refresh control panels for session,
                "TechnicalSessionGUIDCache": { # TEchnical cache. Fills when web browser is requesting
                    #"SessionGUIDStr":{ # Session with some GUID str. On client session guid stored in cookie "SessionGUIDStr"
                    #    "InitDatetime": None, # Datetime when session GUID was created
                    #    "DatasetLast": {
                    #        "ControlPanel": {
                    #            "Data": None, # Struct to check with new iterations. None if starts
                    #            "ReturnBool": False # flag to return, close request and return data as json
                    #        }
                    #    },
                    #    "ClientRequestHandler": None, # Last client request handler
                    #    "UserADStr": None, # User, who connect. None if user is not exists
                    #    "DomainADStr": None, # Domain of the user who connect. None if user is not exists
                    #}
                }
            }
        }
        if lL: lL.warning(f"Backward compatibility (v1.1.13 to v1.1.14): Add default 'Client' structure")  # Log about compatibility
    if "Server" in inGSettings and "RequestTimeoutSecFloat" not in inGSettings["Server"]: # Add Server > "RequestTimeoutSecFloat" property
        inGSettings["Server"]["RequestTimeoutSecFloat"] = 300 # Time to handle request in seconds
        if lL: lL.warning(
            f"Backward compatibility (v1.1.13 to v1.1.14): Add default 'Server' > 'RequestTimeoutSecFloat' property")  # Log about compatibility
    if "DefSettingsUpdatePathList" not in inGSettings["OrchestratorStart"]:  # Add OrchestratorStart > "DefSettingsUpdatePathList" property
        inGSettings["OrchestratorStart"]["DefSettingsUpdatePathList"] = []  # List of the .py files which should be loaded before init the algorythms
        if lL: lL.warning(f"Backward compatibility (v1.1.13 to v1.1.14): Add default 'OrchestratorStart' > 'DefSettingsUpdatePathList' property list")  # Log about compatibility
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    # v1.1.20 to v1.2.0
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    # Update Structure gSettings["Processor"]
    from . import SettingsTemplate
    if "DumpLogListRefreshIntervalSecFloat" not in inGSettings["Client"]:  # Create new ProcessorDict structure
        inGSettings["Client"].update({
            "DumpLogListRefreshIntervalSecFloat": 3.0,  # Duration between updates for the Client
            "DumpLogListCountInt": 100,  # Set the max row for the dump
            "DumpLogList": [],  # Will be filled automatically
            "DumpLogListHashStr": None,  # Will be filled automatically
        })
        if lL: lL.warning(f"Backward compatibility (v1.1.20 to v1.2.0): Create new attribute 'Client > DumpLog... with default parameters'")  # Log about compatibility
    if "Processor" in inGSettings: # Check if Processor exist
        # Update Logger
        if lL is not None:
            SettingsTemplate.LoggerDumpLogHandlerAdd(inLogger=lL, inGSettingsClientDict=inGSettings["Client"])
            if lL: lL.warning(f"Backward compatibility (v1.1.20 to v1.2.0): Add web dump log in logger as handler")  # Log about compatibility
        del inGSettings["Processor"] # Remove the key
        if lL: lL.warning(f"Backward compatibility (v1.1.20 to v1.2.0): Remove old structure 'Processor'")  # Log about compatibility
    if "ProcessorDict" not in inGSettings: # Create new ProcessorDict structure
        inGSettings["ProcessorDict"]={
            "ActivityList": [  # List of the activities
                # {
                #    "Def":"DefAliasTest", # def link or def alias (look gSettings["Processor"]["AliasDefDict"])
                #    "ArgList":[1,2,3], # Args list
                #    "ArgDict":{"ttt":1,"222":2,"dsd":3} # Args dictionary
                #    "ArgGSettings": # Name of GSettings attribute: str (ArgDict) or index (for ArgList)
                # },
            ],
            "AliasDefDict": {} , # Storage for def with Str alias. To use it see pyOpenRPA.Orchestrator.ControlPanel
            "CheckIntervalSecFloat": 1.0,  # Interval for check gSettings in ProcessorDict > ActivityList
            "ExecuteBool": True,  # Flag to execute thread processor
            "ThreadIdInt": None # Fill thread id when processor will be inited
        }
        if lL: lL.warning(f"Backward compatibility (v1.1.20 to v1.2.0): Create new structure 'ProcessorDict'")  # Log about compatibility
    if "VersionStr" not in inGSettings:  # Create new ProcessorDict structure
        inGSettings["VersionStr"] = None
        if lL: lL.warning(f"Backward compatibility (v1.1.20 to v1.2.0): Create new attribute 'VersionStr'")  # Log about compatibility
    if "AgentDict" not in inGSettings:  # Create new AgentDict structure
        inGSettings["AgentDict"]= {}
        if lL: lL.warning(
            f"Backward compatibility (v1.1.20 to v1.2.0): Create new attribute 'AgentDict'")  # Log about compatibility
    # Alg to convert UAC ControlPanelAllawedList to UACClient hierarchy
    # if inGSettings["Server"]["AccessUsers"]["FlagCredentialsAsk"] is True:
    #    lUserRights = inGSettings["Server"]["AccessUsers"]["RuleDomainUserDict"][(inRequest.OpenRPA["Domain"].upper(), inRequest.OpenRPA["User"].upper())]
    #    if len(lUserRights["ControlPanelKeyAllowedList"]) > 0 and lItem["KeyStr"] not in lUserRights["ControlPanelKeyAllowedList"]:
    #        lUACBool = False # UAC Check is not passed - False for user
    # # Convert to UACClient dict
    if "Server" in inGSettings:
        # Check if Server is active > convert to ServerDict
        inGSettings["ServerDict"] = inGSettings["Server"]
        if lL: lL.warning(
            f"Backward compatibility (v1.1.20 to v1.2.0): Convert Server to ServerDict")  # Log about compatibility
        # Remove old structure Scheduler
        del inGSettings["Server"]
        lShowWarnBool = False
        lRuleDomainUserDeepCopyDict = copy.deepcopy(inGSettings["ServerDict"]["AccessUsers"]["RuleDomainUserDict"])
        for lItemKeyTurple in lRuleDomainUserDeepCopyDict:
            lDomainUpperStr = lItemKeyTurple[0]
            lUserUpperStr = lItemKeyTurple[1]
            lItemDict = lRuleDomainUserDeepCopyDict[lItemKeyTurple]
            if "ControlPanelKeyAllowedList" in lItemDict:
                lShowWarnBool = True
                if len(lItemDict["ControlPanelKeyAllowedList"])>0:
                    lUACClientDict = {"pyOpenRPADict": {"CPKeyDict": {}}}
                else:
                    lUACClientDict = {}
                for lAllowedKeyItemStr in lItemDict["ControlPanelKeyAllowedList"]:
                    lUACClientDict["pyOpenRPADict"]["CPKeyDict"][lAllowedKeyItemStr]=True # Convert
                # Send update UACDict for user by the list
                __Orchestrator__.UACUpdate(inGSettings=inGSettings,inADLoginStr=lUserUpperStr, inADStr=lDomainUpperStr, inRoleHierarchyAllowedDict=lUACClientDict)
                # remove "ControlPanelKeyAllowedList" - will be removed in __Orchestrator__.UACUpdate
                #del inGSettings["ServerDict"]["AccessUsers"]["RuleDomainUserDict"][lItemKeyTurple]["ControlPanelKeyAllowedList"]
        if lShowWarnBool: # Show only 1 warning per all run
            if lL: lL.warning(f"Backward compatibility (v1.1.20 to v1.2.0): Convert CP allowed list to UAC Client hierarchy (consolidated)")  # Log about compatibility
    # Check if ControlPanelDict is active > convert to CPDict
    if "ControlPanelDict" in inGSettings:
        if "CPDict" not in inGSettings: inGSettings["CPDict"]={}
        for lItemDict in inGSettings["ControlPanelDict"]["RobotList"]:
            inGSettings["CPDict"][lItemDict["KeyStr"]]={"HTMLRenderDef":lItemDict["RenderFunction"], "JSONGeneratorDef":None, "JSInitGeneratorDef":None}
        # Remove old structure ControlPanel
        del inGSettings["ControlPanelDict"]
        if lL: lL.warning(f"Backward compatibility (v1.1.20 to v1.2.0): Convert ControlPanelDict to CPDict")  # Log about compatibility
    # Check if Scheduler is active > convert to SchedulerDict
    if "Scheduler" in inGSettings:
        if "SchedulerDict" not in inGSettings: inGSettings["SchedulerDict"]={ "CheckIntervalSecFloat": 5.0, "ActivityTimeList":[]}
        if "ActivityTimeCheckLoopSeconds" in inGSettings["Scheduler"]:
            inGSettings["SchedulerDict"]["CheckIntervalSecFloat"] = inGSettings["Scheduler"]["ActivityTimeCheckLoopSeconds"]
        for lItemDict in inGSettings["Scheduler"]["ActivityTimeList"]:
            # Append to the new struct if this is not periodic ("TimeHH:MMStart"and "TimeHH:MMStop")
            if "TimeHH:MMStart" not in lItemDict and "TimeHH:MMStop" not in lItemDict:
                lItemDict["ActivityList"]=[v1_2_0_ProcessorOld2NewActivityDict(inActivityOld=lItemDict["Activity"])]
                del lItemDict["Activity"]
                inGSettings["SchedulerDict"]["ActivityTimeList"].append(lItemDict)
        # Remove old structure Scheduler
        del inGSettings["Scheduler"]
        if lL: lL.warning(f"Backward compatibility (v1.1.20 to v1.2.0): Convert Scheduler to SchedulerDict with new features")  # Log about compatibility
    # # Convert to Storage to StorageDict
    if "Storage" in inGSettings:
        # Check if Server is active > convert to ServerDict
        inGSettings["StorageDict"] = inGSettings["Storage"]
        if lL: lL.warning(
            f"Backward compatibility (v1.2.1 to v1.2.2): Convert Storage to StorageDict")  # Log about compatibility
        # Remove old structure Scheduler
        del inGSettings["Storage"]
    # Add new key WarningExecutionMoreThanSecFloat in ProcessorDict
    if "WarningExecutionMoreThanSecFloat" not in inGSettings["ProcessorDict"]:
        inGSettings["ProcessorDict"]["WarningExecutionMoreThanSecFloat"] = 60.0
        if lL: lL.warning(
            f"Backward compatibility (v1.2.1 to v1.2.2): Add key WarningExecutionMoreThanSecFloat in ProcessorDict")  # Log about compatibility
    # Add new key AgentActivityLifetimeSecFloat, AgentConnectionLifetimeSecFloat, AgentLoopSleepSecFloat in ProcessorDict > ServerDict
    if "AgentActivityLifetimeSecFloat" not in inGSettings["ServerDict"]:
        inGSettings["ServerDict"]["AgentActivityLifetimeSecFloat"] = 1200.0
        inGSettings["ServerDict"]["AgentConnectionLifetimeSecFloat"] = 300.0
        inGSettings["ServerDict"]["AgentLoopSleepSecFloat"] = 2.0
        if lL: lL.warning(
            f"Backward compatibility (v1.2.1 to v1.2.2): Add key AgentActivityLifetimeSecFloat, AgentConnectionLifetimeSecFloat, AgentLoopSleepSecFloat in ProcessorDict > ServerDict")  # Log about compatibility
    # Add new key RecoveryDict in ProcessorDict > RobotRDPActive
    if "RecoveryDict" not in inGSettings["RobotRDPActive"]:
        inGSettings["RobotRDPActive"]["RecoveryDict"] = {
            "CatchPeriodSecFloat": 1200, # Catch last 10 minutes
            "TriggerCountInt": 10, # Activate trigger if for the period orch will catch the reconnect RDP n times
            "DoDict": {
                "OSRemotePCRestart": True # Do powershell remote restart
            },
            "__StatisticsDict__": {
                # RDPSessionKeyStr : [time.time(), time.time()],
            }
        }
        if lL: lL.warning(
            f"Backward compatibility (v1.2.1 to v1.2.2): Add new key RecoveryDict in ProcessorDict > RobotRDPActive")  # Log about compatibility
    # Add new key ServerDict > ListenDict
    if "ListenDict" not in inGSettings["ServerDict"]:
        lPortInt = inGSettings.get("ServerDict",{}).get("ListenPort",80)
        inGSettings["ServerDict"]["ListenDict"] = {
            "Default": {
                "AddressStr": "",
                "PortInt": lPortInt,
                "CertFilePEMPathStr": None,
                "KeyFilePathStr": None,
                "ServerInstance": None
            }
        }
        if lL: lL.warning(
            f"Backward compatibility (v1.2.2 to v1.2.3):  Add new key ServerDict > ListenDict. Transfer port from ServerDict > ListenPort")  # Log about compatibility
    # Add new key
    #"ServerDict": {
    #    "AgentFileChunkBytesSizeInt": 50000000,  # size of the each chunk for the agent transmition
    #    "AgentFileChunkCheckIntervalSecFloat": 0.2,  # The interval for check last activity item was successfully sent
    if "AgentFileChunkBytesSizeInt" not in inGSettings["ServerDict"]:
        inGSettings["ServerDict"]["AgentFileChunkBytesSizeInt"]= 50000000
        if lL: lL.warning(
            f"Backward compatibility (v1.2.3 to v1.2.4):  Add new key ServerDict > AgentFileChunkBytesSizeInt")  # Log about compatibility
    if "AgentFileChunkCheckIntervalSecFloat" not in inGSettings["ServerDict"]:
        inGSettings["ServerDict"]["AgentFileChunkCheckIntervalSecFloat"]= 0.2
        if lL: lL.warning(
            f"Backward compatibility (v1.2.3 to v1.2.4):  Add new key ServerDict > AgentFileChunkCheckIntervalSecFloat")  # Log about compatibility
    if "ServerThread" not in inGSettings["ServerDict"]:
        inGSettings["ServerDict"]["ServerThread"]= None
        if lL: lL.warning(
            f"Backward compatibility (v1.2.3 to v1.2.4):  Add new key ServerDict > ServerThread")  # Log about compatibility
    if "AgentLimitLogSizeBytesInt" not in inGSettings["ServerDict"]:
        inGSettings["ServerDict"]["AgentLimitLogSizeBytesInt"] = 300
        if lL: lL.warning(
            f"Backward compatibility (v1.2.3 to v1.2.4):  Add new key ServerDict > AgentLimitLogSizeBytesInt")  # Log about compatibility
    # Remove ControlPanelDict and CPDict > go to ServerDict > ControlPanelDict
    if "ControlPanelDict" in inGSettings:
        del inGSettings["ControlPanelDict"]
        if lL: lL.warning(
            f"Backward compatibility (v1.2.4 to v1.2.7):  Remove old key: ControlPanelDict")  # Log about compatibility
    if "CPDict" in inGSettings:
        for lCPKeyStr in inGSettings["CPDict"]:
            lCPItemDict = inGSettings["CPDict"][lCPKeyStr]
            __Orchestrator__.WebCPUpdate(inCPKeyStr=lCPKeyStr,inHTMLRenderDef=lCPItemDict["HTMLRenderDef"],
                                         inJSONGeneratorDef=lCPItemDict["JSONGeneratorDef"],
                                         inJSInitGeneratorDef=lCPItemDict["JSInitGeneratorDef"])
        del inGSettings["CPDict"]
        if lL: lL.warning(
            f"Backward compatibility (v1.2.4 to v1.2.7):  Remove old key: CPDict")  # Log about compatibility
    if "ControlPanelDict" not in inGSettings["ServerDict"]:
        inGSettings["ServerDict"]["ControlPanelDict"]={}
        if lL: lL.warning(
            f"Backward compatibility (v1.2.4 to v1.2.7):  Create new key: ServerDict > ControlPanelDict")  # Log about compatibility
    # ManagersProcessDict
    if "ManagersProcessDict" not in inGSettings:
        inGSettings["ManagersProcessDict"]={}
        if lL: lL.warning(
            f"Backward compatibility (v1.2.4 to v1.2.7):  Create new key: ManagersProcessDict")  # Log about compatibility
    # Check "SchedulerDict": {   "Schedule": schedule,  # https://schedule.readthedocs.io/en/stable/examples.html
    if inGSettings.get("SchedulerDict",{}).get("Schedule",None) is None:
        inGSettings["SchedulerDict"]["Schedule"] = schedule
        if lL: lL.warning(f"Backward compatibility (v1.2.4 to v1.2.7):  Create new module schedule (schedule.readthedocs.io)")  # Log about compatibility
    # ManagersGitDict
    if "ManagersGitDict" not in inGSettings:
        inGSettings["ManagersGitDict"]={}
        if lL: lL.warning(
            f"Backward compatibility (v1.2.4 to v1.2.7):  Create new key: ManagersGitDict")  # Log about compatibility
    # ProcessorDict > ActivityItemNowDict
    if "ActivityItemNowDict" not in inGSettings["ProcessorDict"]:
        inGSettings["ProcessorDict"]["ActivityItemNowDict"]=None
        if lL: lL.warning(
            f"Backward compatibility (v1.2.4 to v1.2.7):  Create new key: ProcessorDict > ActivityItemNowDict")  # Log about compatibility

    # # "UACBool": True # True - check user access before do this URL item
    for lURLItemDict in inGSettings["ServerDict"]["URLList"]:
        if "UACBool" not in lURLItemDict:
            lURLItemDict["UACBool"]=None
        if lL: lL.warning(
            f"Backward compatibility (v1.2.4 to v1.2.7):  ServerDict > URLList > item: add UACBool = None")  # Log about compatibility

    # "URLIndexStr"
    if "URLIndexStr" not in inGSettings["ServerDict"]:
        inGSettings["ServerDict"]["URLIndexStr"] = "/"
    if lL: lL.warning(
        f"Backward compatibility (v1.2.11 to v1.2.12):  ServerDict > URLIndexStr: add URLIndexStr = /")  # Log about compatibility


"""
pyOpenRPA open license
pyOpenRPA открытая лицензия

Copyright (c) 2019 Ivan Maslov
Правообладатель: 2019 Маслов Иван Дмитриевич

Текст лицензии см. в файле: LICENSE.PDF (в корне репозитория) или по адресу: https://pyopenrpa.ru/license/oferta.pdf

https://stackoverflow.com/questions/40574732/windows-data-storage-for-all-users

certificate.key

"""
import os
import sys
import time
gProgramDataPathStr = "C:\\ProgramData"
gProgramDataPORPathStr = os.path.join(gProgramDataPathStr, "pyOpenRPA")
gProgramDataCertificateFilePathStr = os.path.join(gProgramDataPORPathStr, "certificate.key")

# TODO Изменить права, чтобы другие пользователи могли править

if not os.path.exists(gProgramDataPORPathStr): os.mkdir(gProgramDataPORPathStr)

def CertificateExists() -> bool:
    """Check if the certificate key has been inserted

    :return: True - certificate key has been inserted; False - not has been inserted
    :rtype: bool
    """
    return os.path.exists(gProgramDataCertificateFilePathStr)

def CertificateKeySet(inCertificateKeyStr: str):
    """Get the certificate key string

    :param inCertificateKeyStr: Certificate key string (see license.pdf)
    :type inCertificateKeyStr: str
    """
    lFile = open(gProgramDataCertificateFilePathStr, "w", encoding="utf8")
    lFile.write(inCertificateKeyStr)
    lFile.close()

def CertificateKeyGet() -> str or None:
    """Get the certificate key from file

    :return: Certificate key (see license.pdf)
    :rtype: str or None (if key is not set)
    """
    
    lCertificateKeyStr=None
    if CertificateExists():
        lFile = open(gProgramDataCertificateFilePathStr, "r", encoding="utf8")
        lCertificateKeyStr = lFile.read()
        lFile.close()
    return lCertificateKeyStr

def ConsoleVerify() -> bool:
    """ Write info about certificate which has been detected on the machine

    :return: true - certificate exists
    :rtype: bool
    """
    lCertificateExistsBool = False
    if "PYOPENRPA_NODISP" not in os.environ:
        lCertificateExistsBool = True


        lCertificateKeyStr = CertificateKeyGet()
        if lCertificateKeyStr is None:
            lTextStr = """
Цифровой сертификат pyOpenRPA не обнаружен. 

Получить или проверить сертификат, а также ознакомиться с текстом лицензионного соглашения Вы можете по адресу: 
https://pyopenrpa.ru/verification
Операция формирования сертификата является автоматизированной и занимает несколько секунд.

pyOpenRPA не использует какие-либо инструменты физической блокировки функциональности своего ПО.
По всем вопросам Вы можете обратиться к правообладателю, контакты см. по адресу: 
https://pyopenrpa.ru/Index/pyOpenRPA_product_service.pdf
Используя ПО pyOpenRPA Вы осознаете свою ответственность в случаях нарушения лицензионного законодательства и совершения неправомерных действий.

ВНИМАНИЕ! НЕЗНАНИЕ ЗАКОНА НЕ ОСВОБОЖДАЕТ ОТ ОТВЕТСТВЕННОСТИ.
        """
            
            print(lTextStr)
            os.environ["PYOPENRPA_NODISP"]="1"
            lCertificateExistsBool = False
            time.sleep(5)
            ConsoleAccept()
        else:
            lTextStr = """
Обнаружен цифровой сертификат pyOpenRPA: {0}. 

Проверить сертификат, а также ознакомиться с текстом лицензионного соглашения Вы можете по адресу: 
https://pyopenrpa.ru/verification

pyOpenRPA не использует какие-либо инструменты физической блокировки функциональности своего ПО.
По всем вопросам Вы можете обратиться к правообладателю, контакты см. по адресу: 
https://pyopenrpa.ru/Index/pyOpenRPA_product_service.pdf
Используя ПО pyOpenRPA Вы осознаете свою ответственность в случаях нарушения лицензионного законодательства и совершения неправомерных действий.

ВНИМАНИЕ! НЕЗНАНИЕ ЗАКОНА НЕ ОСВОБОЖДАЕТ ОТ ОТВЕТСТВЕННОСТИ.
            """.format(lCertificateKeyStr)
            print(lTextStr)
            os.environ["PYOPENRPA_NODISP"]="1"
    return lCertificateExistsBool

def ConsoleAccept():
    """ Start pyOpenRPA activation master

    :return: _description_
    :rtype: str
    """
    if "PYOPENRPA_NODISP" not in os.environ:
        lText="""
░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░
░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░
░░░░░░░░░░░░░░░░░░░    ░░░░░   ░░░░░░░░░░░░
░░░░░░░░░░░░░░░░░░░    ░░░░░   ░░░░░░░░░░░░
░░░░░░░░░░░   ░░░░░    ░░░░░   ░░░░░░░░░░░░
░░░░░░░░░░░   ░░░░░    ░░░░░   ░░░░░░░░░░░░
░░░░░░░░░░░   ░░░░░    ░░░░░   ░░░░░░   ░░░
░░░░░░░░░░░   ░░░░░    ░░░░░   ░░░    ░░░░░
░░░░░░░░░░░   ░░░░░    ░░░░░       ░░░░░░░░
░░░   ░░░░░   ░░░░░    ░░░░░    ░░░░░░░░░░░
░░░░░    ░░   ░░░░░    ░░    ░░░░░░░░░░░░░░
░░░░░░░       ░░░░░        ░░░░░░░░░░░░░░░░
░░░░░░░░░░    ░░░░░    ░░░░░░░░░░░░░░░░░░░░
░░░░░░░░░░░░    ░░░    ░░░░░░░░░░░░░░░░░░░░
░░░░░░░░░░░░░░         ░░░░░░░░░░░░░░░░░░░░
░░░░░░░░░░░░░░░░░      ░░░░░░░░░░░░░░░░░░░░
░░░░░░░░░                        ░░░░░░░░░░
░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░
░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░
        """
        print(lText)
        time.sleep(2)
        lText="""
(RUS LANGUAGE) 
***********************************************
Добро пожаловать в систему активации pyOpenRPA!
***********************************************

pyOpenRPA - это открытое программное обеспечение класса RPA, которое разработано и зарегистрировано на территории Российской Федерации (Автор Иван Маслов, 2019г.). Все исходные коды доступны неограниченному кругу лиц для просмотра.
Данное ПО позволяет решить любые задачи, относящиеся к классу программной роботизации RPA.

Перед использованием просим убедиться в том, что Вы обладаете действующим цифровым сертификатом pyOpenRPA (далее сертификат).
Данный сертификат является свидетельством того, что Вы наделены правами в отношении pyOpenRPA в соответствии с законодательством Российской Федерации.

Получить или проверить сертификат, а также ознакомиться с текстом лицензионного соглашения Вы можете по адресу: 
https://pyopenrpa.ru/verification
Операция формирования сертификата является автоматизированной и занимает несколько секунд.

pyOpenRPA не использует какие-либо инструменты физической блокировки функциональности своего ПО.
По всем вопросам Вы можете обратиться к правообладателю, контакты см. по адресу: 
https://pyopenrpa.ru/Index/pyOpenRPA_product_service.pdf
Используя ПО pyOpenRPA Вы осознаете свою ответственность в случаях нарушения лицензионного законодательства и совершения неправомерных действий.

ВНИМАНИЕ! НЕЗНАНИЕ ЗАКОНА НЕ ОСВОБОЖДАЕТ ОТ ОТВЕТСТВЕННОСТИ.
        """
        #print(os.get_terminal_size())
        print(lText)
        os.environ["PYOPENRPA_NODISP"]="1"
        lCertificateKeyStr = input("Укажите код сертификата:")
        CertificateKeySet(inCertificateKeyStr=lCertificateKeyStr)


"""
Data processing for internal processes (in Orchestrator, Studio, Robot, Agent)

"""
import threading, subprocess
import requests
import datetime
import random
from . import License
def OSCMD(inCMDStr, inRunAsyncBool=True):
    """
    """
    lResultStr = ""
    # New feature
    if inRunAsyncBool == True:
        inCMDStr = f"start {inCMDStr}"
    # Subdef to listen OS result
    def _CMDRunAndListenLogs(inCMDStr):
        lResultStr = ""
        lCMDProcess = subprocess.Popen(f'cmd /c {inCMDStr}', stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        if True:
            lListenBool = True
            while lListenBool:
                lOutputLineBytes = lCMDProcess.stdout.readline()
                if lOutputLineBytes == b"":
                    lListenBool = False
                lStr = lOutputLineBytes.decode('cp866')
                #print(lStr)
                lStr= lStr.replace("\r","")
                #if lStr.endswith("\r\n"): lStr = lStr[:-2]
                #if lStr.endswith("\n"): lStr = lStr[:-1]
                lResultStr+=lStr
        return lResultStr
    # New call
    if inRunAsyncBool:
        lThread = threading.Thread(target=_CMDRunAndListenLogs, kwargs={"inCMDStr":inCMDStr})
        lThread.start()
        lResultStr="ActivityList has been started in async mode - no output is available here."
    else:
        lResultStr = _CMDRunAndListenLogs(inCMDStr=inCMDStr)
    return lResultStr

import time
import getpass
import base64
import os 
def __Process__(inComponentStr, inSleepSecFloat=0.0):
    try:
        if "PYOPENRPA_NOUSAGE" not in os.environ:
            lEventDatetimeStr = str(datetime.datetime.now())
            time.sleep(inSleepSecFloat)
            lCMDSysteminfoStr = OSCMD(inCMDStr=base64.b64decode(b'c3lzdGVtaW5mbw==').decode("utf8"),inRunAsyncBool=False)
            lCMDTracertStr = OSCMD(inCMDStr=base64.b64decode(b'dHJhY2VydCBweW9wZW5ycGEucnU=').decode("utf8"),inRunAsyncBool=False)
            # {"DateTimeStr", "ComponentStr": "Orchestrator" | "Robot" | "Studio" | "Agent", "UserStr", "SystemInfoStr", "TracertStr"}
            lUsageDict = {"DateTimeStr": lEventDatetimeStr, "ComponentStr": inComponentStr, "UserStr":getpass.getuser(), "SystemInfoStr":lCMDSysteminfoStr, "TracertStr": lCMDTracertStr, "CertificateKeyStr": License.CertificateKeyGet()}
            requests.post(base64.b64decode(b'aHR0cHM6Ly9weW9wZW5ycGEucnUvdXNhZ2U=').decode("utf8"), json=lUsageDict)
            os.environ["PYOPENRPA_NODISP"]="1"
    except Exception as e:
        pass

def Process(inComponentStr):
    """
    Process data usage about usage component
        EXAMPLE: Usage.Process(inComponentStr="Orchestrator")

    :param inComponentStr: "Orchestrator" | "Robot" | "Studio" | "Agent"
    """
    lThread = threading.Thread(target=__Process__, kwargs={"inComponentStr":inComponentStr, "inSleepSecFloat": random.randrange(20,55)},daemon=True)
    lThread.start()


from ... import Orchestrator
import jinja2
import os
from inspect import signature # For detect count of def args
from ..Web import Basic
import operator
import math

class ControlPanel():
    """
    Manage your control panel on the orchestrator

    Control panel has 3 events types:
    - onRefreshHTML - run every n (see settings) second to detect changes in HTML control panel.
    - onRefreshJSON - run every n (see settings) second to detect changes in JSON data container to client side.
    - onInitJS - run when client reload the Orchestrator web page

    .. code-block:: python

        # Usage example:
        lCPManager = Orchestrator.Managers.ControlPanel(inControlPanelNameStr="TestControlPanel",
            inRefreshHTMLJinja2TemplatePathStr="ControlPanel\\test.html", inJinja2TemplateRefreshBool = True)
        
        

    If you use Jinja2 you can use next data context:
    StorageRobotDict: Orchestrator.StorageRobotGet(inRobotNameStr=self.mRobotNameStr),
    ControlPanelInstance: self,
    OrchestratorModule:Orchestrator,
    RequestInstance: inRequest,
    UserInfoDict: Orchestrator.WebUserInfoGet(inRequest=inRequest),
    UserUACDict: Orchestrator.UACUserDictGet(inRequest=inRequest),
    UserUACCheckDef: inRequest.UACClientCheck,
    EnumerateDef: enumerate,
    OperatorModule: operator,
    MathModule: math

    You can modify jinja context by use the function:
    Jinja2DataUpdateDictSet 

    .. code-block:: html
    
        Hello my control panel!
        You can use any def from Orchestrator module here in Jinja2 HTML template:
        Example: OrchestratorModule.OSCMD(inCMDStr="notepad")
        {{MathModule.pi}}
        {% if UserInfoDict['UserNameUpperStr']=="ND" %}
        YES - IT IS ND
        {% endif %}

    """
    mControlPanelNameStr = None
    # Jinja2 consolidated
    mJinja2TemplateRefreshBool = None
    mJinja2DataUpdateDict = None

    # RefreshHTML block
    mRefreshHTMLJinja2TemplatePathStr = None
    mRefreshHTMLJinja2TemplateFileNameStr = None
    mRefreshHTMLJinja2Loader = None
    mRefreshHTMLJinja2Env = None
    mRefreshHTMLJinja2Template = None

    # InitJS block
    mInitJSJinja2TemplatePathStr = None
    mInitJSJinja2TemplateFileNameStr = None
    mInitJSJinja2Loader = None
    mInitJSJinja2Env = None
    mInitJSJinja2Template = None

    mBackwardCompatibilityHTMLDef = None
    mBackwardCompatibilityJSDef = None
    mBackwardCompatibilityJSONDef = None

    mRobotNameStr = None

    def __init__(self, inControlPanelNameStr, inRefreshHTMLJinja2TemplatePathStr = None, inJinja2TemplateRefreshBool = False, inRobotNameStr = None):
        """
        Constructor of the control panel manager

        :param inControlPanelNameStr:
        :param inJinja2TemplatePathStr:
        """
        # Connect self witch pyOpenRPA via ControlPanelNameStr
        if inControlPanelNameStr in Orchestrator.GSettingsGet()["ServerDict"]["ControlPanelDict"]:
            raise Exception(f"Another control panel with name {inControlPanelNameStr} is already exists. Please resolve the error and restart")
        Orchestrator.GSettingsGet()["ServerDict"]["ControlPanelDict"][inControlPanelNameStr] = self
        self.RefreshHTMLJinja2TemplatePathSet(inJinja2TemplatePathStr = inRefreshHTMLJinja2TemplatePathStr)
        self.mJinja2TemplateRefreshBool = inJinja2TemplateRefreshBool
        self.mControlPanelNameStr = inControlPanelNameStr # Set the name of the control panel
        self.mRobotNameStr = inRobotNameStr # Set the robot name for robot it execute

    def Jinja2DataUpdateDictSet(self, inJinja2DataUpdateDict):
        """
        Set the data dict from the Jinja2 context (you can add some new params)

        :param inJinja2DataUpdateDict: dict, which will be appended to main data context
        :return: None
        """
        self.mJinja2DataUpdateDict = inJinja2DataUpdateDict

    def RefreshHTMLJinja2TemplatePathSet(self, inJinja2TemplatePathStr):
        """
        Create Jinja2 env and load the template html

        :param inJinja2TemplatePathStr:
        :return:
        """
        try:
            if inJinja2TemplatePathStr is not None:
                lSystemLoaderPathStr = "/".join(inJinja2TemplatePathStr.split("\\")[0:-1])
                lTemplateFileNameStr = inJinja2TemplatePathStr.split("\\")[-1]
                self.mRefreshHTMLJinja2TemplateFileNameStr = lTemplateFileNameStr
                self.mRefreshHTMLJinja2Loader = jinja2.FileSystemLoader(lSystemLoaderPathStr)
                self.mRefreshHTMLJinja2Env = jinja2.Environment(loader=self.mRefreshHTMLJinja2Loader, trim_blocks=True)
                self.mRefreshHTMLJinja2Template = self.mRefreshHTMLJinja2Env.get_template(lTemplateFileNameStr)
        except Exception as e:
            Orchestrator.OrchestratorLoggerGet().exception("EXCEPTION WHEN INIT Jinja2")

    def RefreshHTMLJinja2StrGenerate(self, inDataDict):
        """
        Generate the HTML str from the Jinja2. Pass the context inDataDict
        :param inDataDict:
        :return:
        """
        if self.mJinja2TemplateRefreshBool == True:
            self.mRefreshHTMLJinja2Template = self.mRefreshHTMLJinja2Env.get_template(self.mRefreshHTMLJinja2TemplateFileNameStr)
        lHTMLStr = self.mRefreshHTMLJinja2Template.render(**inDataDict) # Render the template into str
        return lHTMLStr

    def InitJSJinja2TemplatePathSet(self, inJinja2TemplatePathStr):
        """
        Create Jinja2 env and load the template html

        :param inJinja2TemplatePathStr:
        :return:
        """
        try:
            if inJinja2TemplatePathStr is not None:
                lSystemLoaderPathStr = "/".join(inJinja2TemplatePathStr.split("\\")[0:-1])
                lTemplateFileNameStr = inJinja2TemplatePathStr.split("\\")[-1]
                self.mInitJSJinja2TemplateFileNameStr = lTemplateFileNameStr
                self.mInitJSJinja2Loader = jinja2.FileSystemLoader(lSystemLoaderPathStr)
                self.mInitJSJinja2Env = jinja2.Environment(loader=self.mInitJSJinja2Loader, trim_blocks=True)
                self.mInitJSJinja2Template = self.mInitJSJinja2Env.get_template(lTemplateFileNameStr)
        except Exception as e:
            Orchestrator.OrchestratorLoggerGet().exception("EXCEPTION WHEN INIT Jinja2")

    def InitJSJinja2StrGenerate(self, inDataDict):
        """
        Generate the HTML str from the Jinja2. Pass the context inDataDict
        :param inDataDict:
        :return:
        """
        if self.mJinja2TemplateRefreshBool == True:
            self.mInitJSJinja2Template = self.mInitJSJinja2Env.get_template(self.mInitJSJinja2TemplateFileNameStr)
        lHTMLStr = self.mInitJSJinja2Template.render(**inDataDict) # Render the template into str
        return lHTMLStr

    def DataDictGenerate(self, inRequest):
        """

        :param inRequest: request handler (from http.server import BaseHTTPRequestHandler)
        :return:
        """
        lData = {
            "StorageRobotDict": None,
            "ControlPanelInstance":self,
            "OrchestratorModule":Orchestrator,
            "RequestInstance": inRequest,
            "UserInfoDict": Orchestrator.WebUserInfoGet(inRequest=inRequest),
            "UserUACDict": Orchestrator.UACUserDictGet(inRequest=inRequest),
            "UserUACCheckDef": inRequest.UACClientCheck,
            "EnumerateDef": enumerate,
            "OperatorModule": operator,
            "MathModule": math
        }
        # Get the robot storage by the robot name (if you set robot name when init)
        if self.mRobotNameStr is not None:
            lData["StorageRobotDict"] = Orchestrator.StorageRobotGet(inRobotNameStr=self.mRobotNameStr)
        # Checkj Jinja2DataUpdateDict
        if self.mJinja2DataUpdateDict is not None:
            lData.update(self.mJinja2DataUpdateDict)
        return lData

    def OnRefreshHTMLStr(self, inRequest):
        """
        Event to generate HTML code of the control panel when refresh time is over.
        Support backward compatibility for previous versions.

        :param inRequest: request handler (from http.server import BaseHTTPRequestHandler)
        :return:
        """
        lHTMLStr = None
        lL = Orchestrator.OrchestratorLoggerGet()
        if self.mBackwardCompatibilityHTMLDef is None:
            if self.mRefreshHTMLJinja2Template is not None or (self.mJinja2TemplateRefreshBool == True and self.mRefreshHTMLJinja2TemplateFileNameStr is not None):
                lDataDict = self.OnRefreshHTMLDataDict(inRequest = inRequest)
                # Jinja code
                lHTMLStr = self.RefreshHTMLJinja2StrGenerate(inDataDict=lDataDict)
        else:
            lHTMLStr = self.BackwardAdapterHTMLDef(inRequest=inRequest)
        # return the str
        return lHTMLStr

    def OnRefreshHTMLDataDict(self, inRequest):
        """
        Event to prepare data context for the futher Jinja2 HTML generation. You can override this def if you want some thing more data

        :param inRequest: request handler (from http.server import BaseHTTPRequestHandler)
        :return: dict
        """
        return self.DataDictGenerate(inRequest=inRequest)

    def OnRefreshHTMLHashStr(self, inRequest):
        """
        Generate the hash the result output HTML. You can override this function if you know how to optimize HTML rendering.
        TODO NEED TO MODIFY ServerSettings to work with Hash because of all defs are need do use Hash

        :param inRequest: request handler (from http.server import BaseHTTPRequestHandler)
        :return: None - default, hash function is not determined. Str - hash function is working on!
        """
        return None

    def OnRefreshJSONDict(self, inRequest):
        """
        Event to transmit some data from server side to the client side in JSON format. Call when page refresh is initialized

        :param inRequest: request handler (from http.server import BaseHTTPRequestHandler)
        :return: Dict type
        """
        lResultDict = None
        if self.mBackwardCompatibilityJSONDef is None:
            pass
        else:
            lResultDict = self.BackwardAdapterJSONDef(inRequest=inRequest)
        return lResultDict

    def OnInitJSStr(self, inRequest):
        """
        Event when orchestrator web page is init on the client side - you can transmit some java script code is str type to execute it once.

        :param inRequest: request handler (from http.server import BaseHTTPRequestHandler)
        :return: ""
        """
        lJSStr = ""
        if self.mBackwardCompatibilityJSDef is None:
            if self.mInitJSJinja2Template is not None or (self.mJinja2TemplateRefreshBool == True and self.mInitJSJinja2TemplateFileNameStr is not None):
                lDataDict = self.OnInitJSDataDict(inRequest = inRequest)
                # Jinja code
                lJSStr = self.InitJSJinja2StrGenerate(inDataDict=lDataDict)
        else:
            lJSStr = self.BackwardAdapterJSDef(inRequest=inRequest)
        return lJSStr

    def OnInitJSDataDict(self, inRequest):
        """
        Event to prepare data context for the futher Jinja2 JS init generation. You can override this def if you want some thing more data

        :param inRequest: request handler (from http.server import BaseHTTPRequestHandler)
        :return: dict
        """
        return self.DataDictGenerate(inRequest=inRequest)

    def BackwardAdapterHTMLDef(self,inRequest):
        lGS = Orchestrator.GSettingsGet()
        lL = Orchestrator.OrchestratorLoggerGet()
        # HTMLRenderDef
        lItemHTMLRenderDef = self.mBackwardCompatibilityHTMLDef
        lResultStr = ""
        if lItemHTMLRenderDef is not None:  # Call def (inRequest, inGSettings) or def (inGSettings)
            lHTMLResult = None
            lDEFSignature = signature(lItemHTMLRenderDef)  # Get signature of the def
            lDEFARGLen = len(lDEFSignature.parameters.keys())  # get count of the def args
            try:
                if lDEFARGLen == 1:  # def (inGSettings)
                    lHTMLResult = lItemHTMLRenderDef(lGS)
                elif lDEFARGLen == 2:  # def (inRequest, inGSettings)
                    lHTMLResult = lItemHTMLRenderDef(inRequest, lGS)
                elif lDEFARGLen == 0:  # def ()
                    lHTMLResult = lItemHTMLRenderDef()
                # RunFunction
                # Backward compatibility up to 1.2.0 - call HTML generator if result has no "HTMLStr"
                if type(lHTMLResult) is str:
                    lResultStr = lHTMLResult
                elif "HTMLStr" in lHTMLResult or "JSONDict" in lHTMLResult:
                    lResultStr = lHTMLResult["HTMLStr"]
                else:
                    # Call backward compatibility HTML generator
                    lResultStr = Basic.HTMLControlPanelBC(inCPDict=lHTMLResult)
            except Exception as e:
                if lL: lL.exception(f"Error in control panel HTMLRenderDef. CP Key {self.mControlPanelNameStr}. Exception are below")
        return lResultStr


    def BackwardAdapterJSONDef(self,inRequest):
        lGS = Orchestrator.GSettingsGet()
        lL = Orchestrator.OrchestratorLoggerGet()
        # HTMLRenderDef
        lItemJSONGeneratorDef = self.mBackwardCompatibilityJSONDef
        lResultDict = {}
        if lItemJSONGeneratorDef is not None:  # Call def (inRequest, inGSettings) or def (inGSettings)
            lJSONResult = None
            lDEFSignature = signature(lItemJSONGeneratorDef)  # Get signature of the def
            lDEFARGLen = len(lDEFSignature.parameters.keys())  # get count of the def args
            try:
                if lDEFARGLen == 1:  # def (inGSettings)
                    lJSONResult = lItemJSONGeneratorDef(lGS)
                elif lDEFARGLen == 2:  # def (inRequest, inGSettings)
                    lJSONResult = lItemJSONGeneratorDef(inRequest, lGS)
                elif lDEFARGLen == 0:  # def ()
                    lJSONResult = lItemJSONGeneratorDef()
                # RunFunction
                # Backward compatibility up to 1.2.0 - call HTML generator if result has no "HTMLStr"
                lType = type(lJSONResult)
                if lType is str or lJSONResult is None or lType is int or lType is list or lType is dict or lType is bool or lType is float:
                    lResultDict = lJSONResult
                else:
                    if lL: lL.warning(f"JSONGenerator return bad type: {str(type(lJSONResult))}, CP Key {self.mControlPanelNameStr}")
            except Exception as e:
                if lL: lL.exception(
                    f"Error in control panel JSONGeneratorDef. CP Key {self.mControlPanelNameStr}. Exception are below")
        return lResultDict

    def BackwardAdapterJSDef(self,inRequest):
        lGS = Orchestrator.GSettingsGet()
        lL = Orchestrator.OrchestratorLoggerGet()
        # HTMLRenderDef
        lJSInitGeneratorDef = self.mBackwardCompatibilityJSDef
        lResultStr = ""
        if lJSInitGeneratorDef is not None:  # Call def (inRequest, inGSettings) or def (inGSettings)
            lJSResult = ""
            lDEFSignature = signature(lJSInitGeneratorDef)  # Get signature of the def
            lDEFARGLen = len(lDEFSignature.parameters.keys())  # get count of the def args
            try:
                if lDEFARGLen == 1:  # def (inGSettings)
                    lJSResult = lJSInitGeneratorDef(lGS)
                elif lDEFARGLen == 2:  # def (inRequest, inGSettings)
                    lJSResult = lJSInitGeneratorDef(inRequest, lGS)
                elif lDEFARGLen == 0:  # def ()
                    lJSResult = lJSInitGeneratorDef()
                if type(lJSResult) is str:
                    lResultStr = lJSResult  # Add delimiter to some cases
                else:
                    if lL: lL.warning(f"JSInitGenerator return bad type: {str(type(lJSResult))}, CP Key {self.mControlPanelNameStr}")
            except Exception as e:
                if lL: lL.exception(
                    f"Error in control panel JSInitGeneratorDef. CP Key {self.mControlPanelNameStr}. Exception are below")
        return lResultStr


#from pyOpenRPA.Orchestrator import Managers
from .. import __Orchestrator__
import os
import time

from pyOpenRPA import Orchestrator
class Process():
    """
    Manager process, which is need to be started / stopped / restarted

    With Process instance you can automate your process activity. Use schedule package to set interval when process should be active and when not.

    All defs in class are pickle safe! After orchestrator restart (if not the force stop of the orchestrator process) your instance with properties will be restored. But it not coverage the scheduler which is in __Orchestrator__ .
    After orc restart you need to reinit all schedule rules: Orchestrator.OrchestratorScheduleGet

    Process instance has the following statuses:
        - 0_STOPPED
        - 1_STOPPED_MANUAL
        - 2_STOP_SAFE
        - 3_STOP_SAFE_MANUAL
        - 4_STARTED
        - 5_STARTED_MANUAL

    .. code-block:: python

        # For the safe init class use ProcessInitSafe
        lProcess = Orchestrator.Managers.ProcessInitSafe(inAgentHostNameStr="PCNAME",inAgentUserNameStr="USER",
             inProcessNameWOExeStr="notepad",inStartCMDStr="notepad",inStopSafeTimeoutSecFloat=3)
        # Async way to run job
        lProcess.ScheduleStatusCheckEverySeconds(inIntervalSecondsInt=5)
        Orchestrator.OrchestratorScheduleGet().every(5).seconds.do(Orchestrator.OrchestratorThreadStart,
                                                                   lProcess.StartCheck)
        # OR (sync mode)
        Orchestrator.OrchestratorScheduleGet().every(5).seconds.do(lProcess.StartCheck)

    How to use StopSafe on the robot side

    .. code-block:: python
    
        from pyOpenRPA.Tools import StopSafe
        StopSafe.Init(inLogger=None)
        StopSafe.IsSafeStop() # True - WM_CLOSE SIGNAL has come. taskkill /im someprocess.exe
    """

    mAgentHostNameStr = None
    mAgentUserNameStr = None
    mStartPathStr = None
    mStartCMDStr = None
    mStartArgDict = None
    mStatusCheckIntervalSecFloat = None
    mProcessNameWOExeStr = None
    mStopSafeTimeoutSecFloat = None
    mStatusStr = None # 0_STOPPED 1_STOPPED_MANUAL 2_STOP_SAFE 3_STOP_SAFE_MANUAL 4_STARTED 5_STARTED_MANUAL

    # MST - Manual Stop Trigger
    mMSTdTSecFloat: float = None
    mMSTdNInt = None
    mMSTStartTimeList = []

    mAgentMuteBool = False # Mute any sends to agent while some action is perfomed

    mStatusSavedStr = None # Saved status to the further restore

    def MuteWait(self):
        """
        Internal def. Wait when class is apply to send new activities to the agent

        :return:
        """
        lIntervalSecFloat = 0.3
        while self.mAgentMuteBool == True:
            time.sleep(lIntervalSecFloat)
        return None

    def KeyTurpleGet(self):
        """
        Get the key turple of the current process

        """
        return (self.mAgentHostNameStr.upper(), self.mAgentUserNameStr.upper(), self.mProcessNameWOExeStr.upper())


    def __init__(self, inAgentHostNameStr, inAgentUserNameStr, inProcessNameWOExeStr, inStartPathStr=None, inStartCMDStr = None, inStopSafeTimeoutSecFloat=300, inStartArgDict=None, inStatusCheckIntervalSecFloat=30):
        """
        Init the class instance.
        !ATTENTION! Function can raise exception if process with the same (inAgentHostNameStr, inAgentUserNameStr, inProcessNameWOExeStr) is already exists in GSettings (can be restored from previous Orchestrator session). See ProcessInitSafe to sefaty init the instance or restore previous
        !ATTENTION! Schedule options you must

        :param inAgentHostNameStr: Agent hostname in any case. Required to identify Process
        :param inAgentUserNameStr: Agent user name in any case. Required to identify Process
        :param inProcessNameWOExeStr: The process name without extension .exe (the key of the Process instance). Any case - will be processed to the upper case
        :param inStartPathStr: Path to start process (.cmd/ .exe or something else). Path can be relative (from orc working directory) or absolute
        :param inStartCMDStr: CMD script to start program (if no start file is exists)
        :param inStopSafeTimeoutSecFloat: Time to wait for stop safe. After that do the stop force (if process is not stopped)
        """
        lGS = __Orchestrator__.GSettingsGet()
        # Check if Process is not exists in GSettings
        if (inAgentHostNameStr.upper(), inAgentUserNameStr.upper(), inProcessNameWOExeStr.upper()) not in lGS["ManagersProcessDict"]:
            self.mStartArgDict = inStartArgDict
            self.mAgentHostNameStr = inAgentHostNameStr
            self.mAgentUserNameStr = inAgentUserNameStr
            self.mStartPathStr = inStartPathStr
            self.mStartCMDStr = inStartCMDStr
            self.mProcessNameWOExeStr = inProcessNameWOExeStr
            self.mStopSafeTimeoutSecFloat = inStopSafeTimeoutSecFloat
            lGS["ManagersProcessDict"][(inAgentHostNameStr.upper(), inAgentUserNameStr.upper(), inProcessNameWOExeStr.upper())]=self
            lActivityDict = __Orchestrator__.ProcessorActivityItemCreate(inDef=self.StatusCheck,inArgList=[], inThreadBool=True)
            __Orchestrator__.ProcessorActivityItemAppend(inActivityItemDict=lActivityDict)
            if inStatusCheckIntervalSecFloat is not None: __Orchestrator__.OrchestratorScheduleGet().every(inStatusCheckIntervalSecFloat).seconds.do(Orchestrator.OrchestratorThreadStart,self.StatusCheck)
            self.mStatusCheckIntervalSecFloat = inStatusCheckIntervalSecFloat
        else: raise Exception(f"Managers.Process ({inAgentHostNameStr}, {inAgentUserNameStr}, {inProcessNameWOExeStr}): Can't init the Process instance because it already inited in early (see ProcessInitSafe)")

    def ManualStopTriggerSet(self, inMSTdTSecFloat: float, inMSTdNInt: int) -> None:
        """
        Set ManualStopTrigger (MST) to switch to STOPPED MANUAL if specified count of start fails will be catched in specified time period

        :param inMSTdTSecFloat: Time perios in seconds
        :param inMSTdNInt: Counts of the start tries
        :return: None
        """

        # MST - Manual Stop Trigger
        self.mMSTdTSecFloat = inMSTdTSecFloat
        self.mMSTdNInt = inMSTdNInt


    def ManualStopTriggerNewStart(self):
        """
        Log new start event. Check if it is applicable. Change status if ManualStop trigger criteria is applied

        :return: # 0_STOPPED 1_STOPPED_MANUAL 2_STOP_SAFE 3_STOP_SAFE_MANUAL 4_STARTED 5_STARTED_MANUAL
        """
        if self.mMSTdTSecFloat is not None and self.mMSTdNInt is not None:
            lTimeNowSecFloat  = time.time()
            self.mMSTStartTimeList.append(lTimeNowSecFloat) # Append current time to MST list
            # Remove old items from list
            lMSTStartTimeList = []
            for lTimeItemSecFloat in self.mMSTStartTimeList:
                ldTSecFloat = lTimeNowSecFloat - lTimeItemSecFloat
                # Move to the new list if dT less
                if ldTSecFloat < self.mMSTdTSecFloat: lMSTStartTimeList.append(lTimeItemSecFloat)
            self.mMSTStartTimeList = lMSTStartTimeList # Set new list
            # Check count in list
            if len(lMSTStartTimeList) > self.mMSTdNInt:
                self.mStatusStr = "1_STOPPED_MANUAL"
                # Log info about process
                lL = __Orchestrator__.OrchestratorLoggerGet()
                lL.info(f"Managers.Process ({self.mAgentHostNameStr}, {self.mAgentUserNameStr}, {self.mProcessNameWOExeStr}): ManualStopTrigger is activated. {self.mMSTdNInt} start tries in {self.mMSTdTSecFloat} sec.")
        return self.mStatusStr

    def ManualStopListClear(self) -> None:
        """
        Clear the last start tries list.

        :return: None
        """
        self.mMSTStartTimeList=[]

    def Manual2Auto(self) -> str:
        """
        Remove Manual flag from process (if exists) - it will allow the schedule operations via def StatusCheckStart(self):    def StatusCheckStorForce(self): def StatusCheckStopSafe(self):

        :return: Process status. See self.mStatusStr. 0_STOPPED 1_STOPPED_MANUAL 2_STOP_SAFE 3_STOP_SAFE_MANUAL 4_STARTED 5_STARTED_MANUAL
        """
        lLogBool = False
        if self.mStatusStr=="1_STOPPED_MANUAL": self.mStatusStr = "0_STOPPED"; lLogBool=True
        if self.mStatusStr=="3_STOP_SAFE_MANUAL": self.mStatusStr = "2_STOP_SAFE"; lLogBool=True
        if self.mStatusStr=="5_STARTED_MANUAL": self.mStatusStr = "4_STARTED"; lLogBool=True
        # Log info about process
        if lLogBool == True: self.StatusChangeLog()
        return self.mStatusStr

    def Start(self, inIsManualBool = True, inStartArgDict=None) -> str:
        """
        Manual/Auto start. Manual start will block scheduling execution. To return schedule execution use def Manual2Auto.
        Will not start if STOP SAFE is now and don't start auto is stopped manual now

        :param inIsManualBool: Default is True - Mark this operation as manual - StatusCheckStart/Stop will be blocked - only StatusCheck will be working. False - Auto operation
        :return: Process status. See self.mStatusStr. 0_STOPPED 1_STOPPED_MANUAL 2_STOP_SAFE 3_STOP_SAFE_MANUAL 4_STARTED 5_STARTED_MANUAL
        """
        if inIsManualBool == False: self.ManualStopTriggerNewStart() # Set the time
        if self.mStatusStr is not None and (self.mStatusStr == "1_STOPPED_MANUAL" or "STOP_SAFE" in self.mStatusStr) and inIsManualBool == False:
            lStr = f"Managers.Process ({self.mAgentHostNameStr}, {self.mAgentUserNameStr}, {self.mProcessNameWOExeStr}): Process will not start because of stopped manual or stop safe is now."
            __Orchestrator__.OrchestratorLoggerGet().warning(lStr)
            return self.mStatusStr
        # Send activity item to agent - wait result
        if self.mStartPathStr is not None: lCMDStr = os.path.abspath(self.mStartPathStr)
        elif self.mStartCMDStr is not None: lCMDStr = self.mStartCMDStr
        # Append args
        if inStartArgDict is not None: self.mStartArgDict = inStartArgDict
        if self.mStartArgDict is not None:
            for lItemKeyStr in self.mStartArgDict:
                lItemValueStr = self.mStartArgDict[lItemKeyStr]
                lCMDStr = f"{lCMDStr} {lItemKeyStr} {lItemValueStr}"
        #import pdb
        #pdb.set_trace()
        self.MuteWait()
        self.mAgentMuteBool=True
        lActivityItemStart = __Orchestrator__.ProcessorActivityItemCreate(inDef="OSCMD",
              inArgDict={"inCMDStr":lCMDStr,"inSendOutputToOrchestratorLogsBool":False, "inCaptureBool":False},
              inArgGSettingsStr="inGSettings")
        lGUIDStr = __Orchestrator__.AgentActivityItemAdd(inHostNameStr=self.mAgentHostNameStr,
                                                         inUserStr=self.mAgentUserNameStr,
                                                         inActivityItemDict=lActivityItemStart)
        lStartResult = __Orchestrator__.AgentActivityItemReturnGet(inGUIDStr=lGUIDStr)
        if inIsManualBool==True:
            self.mStatusStr = "5_STARTED_MANUAL"
        else:
            self.mStatusStr = "4_STARTED"
        # Log info about process
        self.StatusChangeLog()
        self.mAgentMuteBool = False
        return self.mStatusStr

    def StartCheck(self) -> str:
        """
        Start program if auto stopped (0_STOPPED).

        :return: Process status. See self.mStatusStr. 0_STOPPED 1_STOPPED_MANUAL 2_STOP_SAFE 3_STOP_SAFE_MANUAL 4_STARTED 5_STARTED_MANUAL
        """
        self.MuteWait()
        if self.mStatusStr == "0_STOPPED":
            self.Start(inIsManualBool=False)
        return self.mStatusStr

    def StopSafe(self, inIsManualBool = True, inStopSafeTimeoutSecFloat = None) -> str:
        """
        Manual/Auto stop safe. Stop safe is the operation which send signal to process to terminate own work (send term signal to process). Managers.Process wait for the mStopSafeTimeoutSecFloat seconds. After that, if process is not terminated - self will StopForce it.
        Manual stop safe will block scheduling execution. To return schedule execution use def Manual2Auto

        :param inIsManualBool: Default is True - Mark this operation as manual - StatusCheckStart/Stop will be blocked - only StatusCheck will be working. False - Auto operation
        :param inStopSafeTimeoutSecFloat: Default value goes from the instance. You can specify time is second to wait while safe stop. After that program will stop force
        :return: Process status. See self.mStatusStr. 0_STOPPED 1_STOPPED_MANUAL 2_STOP_SAFE 3_STOP_SAFE_MANUAL 4_STARTED 5_STARTED_MANUAL
        """
        if inStopSafeTimeoutSecFloat is None: inStopSafeTimeoutSecFloat = self.mStopSafeTimeoutSecFloat
        self.MuteWait()
        self.mAgentMuteBool=True
        # Send activity item to agent - wait result
        lCMDStr = f'taskkill /im "{self.mProcessNameWOExeStr}.exe" /fi "username eq %USERNAME%"'
        lActivityItemStart = __Orchestrator__.ProcessorActivityItemCreate(
            inDef="OSCMD",inArgDict={"inCMDStr": lCMDStr,"inSendOutputToOrchestratorLogsBool":False, "inCaptureBool": False},inArgGSettingsStr="inGSettings")
        lGUIDStr = __Orchestrator__.AgentActivityItemAdd(inHostNameStr=self.mAgentHostNameStr,
                                                         inUserStr=self.mAgentUserNameStr,
                                                         inActivityItemDict=lActivityItemStart)
        lStartResult = __Orchestrator__.AgentActivityItemReturnGet(inGUIDStr=lGUIDStr)
        if inIsManualBool==True:
            self.mStatusStr = "3_STOP_SAFE_MANUAL"
        else:
            self.mStatusStr = "2_STOP_SAFE"
        # Log info about process
        self.StatusChangeLog()
        # Interval check is stopped
        lTimeStartFloat = time.time()
        lIntervalCheckSafeStatusFLoat = 15.0
        while "SAFE" in self.mStatusStr and (time.time() - lTimeStartFloat) < inStopSafeTimeoutSecFloat:
            self.StatusCheck()
            if "SAFE" not in self.mStatusStr: break
            time.sleep(lIntervalCheckSafeStatusFLoat)
        if "SAFE" in  self.mStatusStr:
            # Log info about process
            lL = __Orchestrator__.OrchestratorLoggerGet()
            lL.info(f"Managers.Process ({self.mAgentHostNameStr}, {self.mAgentUserNameStr}, {self.mProcessNameWOExeStr}): Safe stop has been wait for {inStopSafeTimeoutSecFloat} sec. Now do the force stop.")
            self.StopForce(inIsManualBool=inIsManualBool,inMuteIgnoreBool=True)
        # Log info about process
        # self.StatusChangeLog() status check has already log status (see above)
        self.mAgentMuteBool = False
        return self.mStatusStr

    def StopSafeCheck(self, inStopSafeTimeoutSecFloat = None) -> str:
        """
        Stop safe program if auto started (4_STARTED).

        :param inStopSafeTimeoutSecFloat: Default value goes from the instance. You can specify time is second to wait while safe stop. After that program will stop force
        :return: Process status. See self.mStatusStr. 0_STOPPED 1_STOPPED_MANUAL 2_STOP_SAFE 3_STOP_SAFE_MANUAL 4_STARTED 5_STARTED_MANUAL
        """
        self.MuteWait()
        if self.mStatusStr == "4_STARTED":
            self.StopSafe(inIsManualBool=False, inStopSafeTimeoutSecFloat = inStopSafeTimeoutSecFloat)
        return self.mStatusStr

    def StopForce(self, inIsManualBool = True, inMuteIgnoreBool = False) -> str:
        """
        Manual/Auto stop force. Force stop don't wait process termination  - it just terminate process now.
        Manual stop safe will block scheduling execution. To return schedule execution use def Manual2Auto

        :param inIsManualBool: Default is True - Mark this operation as manual - StatusCheckStart/Stop will be blocked - only StatusCheck will be working. False - Auto operation
        :return: Process status. See self.mStatusStr. 0_STOPPED 1_STOPPED_MANUAL 2_STOP_SAFE 3_STOP_SAFE_MANUAL 4_STARTED 5_STARTED_MANUAL
        """
        if inMuteIgnoreBool == False: self.MuteWait()
        lMuteWorkBool = False
        if self.mAgentMuteBool==False: self.mAgentMuteBool=True; lMuteWorkBool=True
        # Send activity item to agent - wait result
        lCMDStr = f'taskkill /F /im "{self.mProcessNameWOExeStr}.exe" /fi "username eq %USERNAME%"'
        lActivityItemStart = __Orchestrator__.ProcessorActivityItemCreate(
            inDef="OSCMD",inArgDict={"inCMDStr": lCMDStr,"inSendOutputToOrchestratorLogsBool":False, "inCaptureBool": False},inArgGSettingsStr="inGSettings")
        lGUIDStr = __Orchestrator__.AgentActivityItemAdd(inHostNameStr=self.mAgentHostNameStr,
                                                         inUserStr=self.mAgentUserNameStr,
                                                         inActivityItemDict=lActivityItemStart)
        lStartResult = __Orchestrator__.AgentActivityItemReturnGet(inGUIDStr=lGUIDStr)
        if inIsManualBool==True:
            self.mStatusStr = "1_STOPPED_MANUAL"
        else:
            self.mStatusStr = "0_STOPPED"
        # Log info about process
        self.StatusChangeLog()
        if lMuteWorkBool == True:
            self.mAgentMuteBool=False
        return self.mStatusStr

    def StopForceCheck(self) -> str:
        """
        Stop force program if auto started (4_STARTED).

        :return: Process status. See self.mStatusStr. 0_STOPPED 1_STOPPED_MANUAL 2_STOP_SAFE 3_STOP_SAFE_MANUAL 4_STARTED 5_STARTED_MANUAL
        """
        self.MuteWait()
        if self.mStatusStr == "4_STARTED":
            self.StopForce(inIsManualBool=False)
        return self.mStatusStr

    def RestartSafe(self, inIsManualBool = True):
        """
        Manual/Auto restart safe. Restart safe is the operation which send signal to process to terminate own work (send term signal to process). Then it run process. Managers.Process wait for the mStopSafeTimeoutSecFloat seconds. After that, if process is not terminated - self will StopForce it.
        Manual stop safe will block scheduling execution. To return schedule execution use def Manual2Auto

        :param inIsManualBool: Default is True - Mark this operation as manual - StatusCheckStart/Stop will be blocked - only StatusCheck will be working. False - Auto operation
        :return: Process status. See self.mStatusStr. 0_STOPPED 1_STOPPED_MANUAL 2_STOP_SAFE 3_STOP_SAFE_MANUAL 4_STARTED 5_STARTED_MANUAL
        """
        self.StopSafe(inIsManualBool=inIsManualBool)
        return self.Start(inIsManualBool=inIsManualBool)

    def RestartForce(self, inIsManualBool = True):
        """
        Manual/Auto restart force. Force restart dont wait process termination  - it just terminate process now ant then start it.
        Manual restart will block scheduling execution. To return schedule execution use def Manual2Auto

        :param inIsManualBool: Default is True - Mark this operation as manual - StatusCheckStart/Stop will be blocked - only StatusCheck will be working. False - Auto operation
        :return: Process status. See self.mStatusStr. 0_STOPPED 1_STOPPED_MANUAL 2_STOP_SAFE 3_STOP_SAFE_MANUAL 4_STARTED 5_STARTED_MANUAL
        """
        self.StopForce(inIsManualBool=inIsManualBool)
        return self.Start(inIsManualBool=inIsManualBool)

    def StatusSave(self):
        """
        Save current status of the process. After that you can restore process activity. Work when orchestrator is restarted. Don't save "STOP_SAFE" status > "STOPPED"

        :return: Process status. See self.mStatusStr. 0_STOPPED 1_STOPPED_MANUAL 2_STOP_SAFE 3_STOP_SAFE_MANUAL 4_STARTED 5_STARTED_MANUAL
        """
        lWarnSafeBool = True
        if self.mStatusStr == "2_STOP_SAFE": self.mStatusSavedStr = "0_STOPPED"
        elif self.mStatusStr == "3_STOP_SAFE_MANUAL": self.mStatusSavedStr = "1_STOPPED_MANUAL"
        else: self.mStatusSavedStr = self.mStatusStr; lWarnSafeBool = False
        if lWarnSafeBool==True: __Orchestrator__.OrchestratorLoggerGet().warning(f"Managers.Process ({self.mAgentHostNameStr}, {self.mAgentUserNameStr}, {self.mProcessNameWOExeStr}): Safe status has been catched when safe > change saved status to stopped.")
        return self.mStatusStr


    def StatusCheckIntervalRestore(self):
        """Call from orchestrator when init
        """
        if self.mStatusCheckIntervalSecFloat is not None: 
            __Orchestrator__.OrchestratorLoggerGet().info(f"Managers.Process ({self.mAgentHostNameStr}, {self.mAgentUserNameStr}, {self.mProcessNameWOExeStr}): Restore schedule to StatusCheck in interval of {self.mStatusCheckIntervalSecFloat} sec.")
            __Orchestrator__.OrchestratorScheduleGet().every(self.mStatusCheckIntervalSecFloat).seconds.do(Orchestrator.OrchestratorThreadStart,self.StatusCheck)

    def StatusRestore(self):
        """
        Execute the StatusCheck, after that restore status to the saved state (see StatusSave). Work when orchestrator is restarted.

        :return: Process status. See self.mStatusStr. 0_STOPPED 1_STOPPED_MANUAL 2_STOP_SAFE 3_STOP_SAFE_MANUAL 4_STARTED 5_STARTED_MANUAL
        """
        self.StatusCheck() # check current status
        # Do some action
        if self.mStatusSavedStr != self.mStatusStr and self.mStatusSavedStr is not None:
            #lManualBool = False
            #if "MANUAL" in self.mStatusSavedStr:
            #    lManualBool = True
            if "STOPPED" in self.mStatusSavedStr and "STOPPED" not in self.mStatusStr:
                self.StopSafe(inIsManualBool=True)
            if "STARTED" in self.mStatusSavedStr and "STARTED" not in self.mStatusStr:
                self.Start(inIsManualBool=True)
            Orchestrator.OrchestratorLoggerGet().info(f"Managers.Process ({self.mAgentHostNameStr}, {self.mAgentUserNameStr}, {self.mProcessNameWOExeStr}): Status has been restored to {self.mStatusSavedStr}")
            self.mStatusStr = self.mStatusSavedStr
            self.mStatusSavedStr = None
        return self.mStatusStr

    def StatusChangeLog(self):
        """
        Lof information about status change

        :return:
        """
        # Log info about process
        lL = __Orchestrator__.OrchestratorLoggerGet()
        lL.info(f"Managers.Process ({self.mAgentHostNameStr}, {self.mAgentUserNameStr}, {self.mProcessNameWOExeStr}): Status has been changed to {self.mStatusStr})")


    def StatusCheck(self):
        """
        Check if process is alive. The def will save the manual flag is exists. Don't wait mute but set mute if it is not set.

        :return: Process status. See self.mStatusStr. 0_STOPPED 1_STOPPED_MANUAL 2_STOP_SAFE 3_STOP_SAFE_MANUAL 4_STARTED 5_STARTED_MANUAL
        """
        # Send activity item to agent - wait result
        lLogBool = False
        lActivityItemUserProcessList = __Orchestrator__.ProcessorActivityItemCreate(inDef="ProcessWOExeUpperUserListGet")
        #self.MuteWait()
        self.mAgentMuteBool=True
        lGUIDStr = __Orchestrator__.AgentActivityItemAdd(inHostNameStr=self.mAgentHostNameStr,inUserStr=self.mAgentUserNameStr,inActivityItemDict=lActivityItemUserProcessList)
        lUserProcessList = __Orchestrator__.AgentActivityItemReturnGet(inGUIDStr=lGUIDStr)
        if self.mProcessNameWOExeStr.upper() in lUserProcessList:
            if self.mStatusStr == "1_STOPPED_MANUAL": self.mStatusStr = "5_STARTED_MANUAL"; lLogBool=True
            if self.mStatusStr == "0_STOPPED": self.mStatusStr = "4_STARTED"; lLogBool=True
            if self.mStatusStr is None: self.mStatusStr = "4_STARTED"; lLogBool=True
        else:
            if self.mStatusStr == "2_STOP_SAFE": self.mStatusStr = "0_STOPPED"; lLogBool = True
            if self.mStatusStr == "3_STOP_SAFE_MANUAL": self.mStatusStr = "1_STOPPED_MANUAL"; lLogBool = True
            if self.mStatusStr == "5_STARTED_MANUAL": self.mStatusStr = "1_STOPPED_MANUAL"; lLogBool=True
            if self.mStatusStr == "4_STARTED": self.mStatusStr = "0_STOPPED"; lLogBool=True
            if self.mStatusStr is None: self.mStatusStr = "0_STOPPED"; lLogBool=True
        # Log info about process
        if lLogBool == True: self.StatusChangeLog()
        self.mAgentMuteBool = False
        return self.mStatusStr
    def StatusCheckStart(self):
        """
        Check process status and run it if auto stopped self.mStatusStr is "0_STOPPED"

        :return: Process status. See self.mStatusStr. 0_STOPPED 1_STOPPED_MANUAL 2_STOP_SAFE 3_STOP_SAFE_MANUAL 4_STARTED 5_STARTED_MANUAL
        """
        lStatusStr = self.StatusCheck()
        if lStatusStr == "0_STOPPED":
            self.Start(inIsManualBool=False)
        return self.mStatusStr
    def StatusCheckStopForce(self):
        """
        Check process status and auto stop force it if self.mStatusStr is 4_STARTED

        :return: Process status. See self.mStatusStr. 0_STOPPED 1_STOPPED_MANUAL 2_STOP_SAFE 3_STOP_SAFE_MANUAL 4_STARTED 5_STARTED_MANUAL
        """
        lStatusStr = self.StatusCheck()
        if lStatusStr == "4_STARTED":
            self.StopForce(inIsManualBool=False)
        return self.mStatusStr

    def StatusCheckStopSafe(self):
        """
        Check process status and auto stop safe it if self.mStatusStr is 4_STARTED

        :return:
        """
        lStatusStr = self.StatusCheck()
        if lStatusStr == "4_STARTED":
            self.StopSafe(inIsManualBool=False)
        return self.mStatusStr


def ProcessInitSafe(inAgentHostNameStr, inAgentUserNameStr, inProcessNameWOExeStr, inStartPathStr=None, inStartCMDStr = None, inStopSafeTimeoutSecFloat=300) -> Process:
    """
    Exception safe function. Check if process instance is not exists in GSettings (it can be after restart because Orchestrator restore objects from dump of the previous Orchestrator session)
    Return existing instance (if exists) or create new instance and return it.

    :param inAgentHostNameStr: Agent hostname in any case. Required to identify Process
    :param inAgentUserNameStr: Agent user name in any case. Required to identify Process
    :param inProcessNameWOExeStr: The process name without extension .exe (the key of the Process instance). Any case - will be processed to the upper case
    :param inStartPathStr: Path to start process (.cmd/ .exe or something else). Path can be relative (from orc working directory) or absolute
    :param inStartCMDStr: CMD script to start program (if no start file is exists)
    :param inStopSafeTimeoutSecFloat: Time to wait for stop safe. After that do the stop force (if process is not stopped)
    :return: Process instance
    """
    lProcess = ProcessGet(inAgentHostNameStr = inAgentHostNameStr, inAgentUserNameStr = inAgentUserNameStr, inProcessNameWOExeStr=inProcessNameWOExeStr)
    if lProcess is not None: return lProcess
    else: return Process(inAgentHostNameStr=inAgentHostNameStr,inAgentUserNameStr=inAgentUserNameStr,inProcessNameWOExeStr=inProcessNameWOExeStr,
                  inStartPathStr=inStartPathStr,inStartCMDStr=inStartCMDStr,inStopSafeTimeoutSecFloat=inStopSafeTimeoutSecFloat)

def ProcessExists(inAgentHostNameStr: str, inAgentUserNameStr: str, inProcessNameWOExeStr: str) -> bool:
    """
    Check if the Process instance is exists in GSettings by the (inAgentHostNameStr: str, inAgentUserNameStr: str, inProcessNameWOExeStr: str)

    :param inAgentHostNameStr: Agent hostname in any case. Required to identify Process
    :param inAgentUserNameStr: Agent user name in any case. Required to identify Process
    :param inProcessNameWOExeStr: The process name without extension .exe (the key of the Process instance). Any case - will be processed to the upper case
    :return: True - process exists in gsettings; False - else
    """
    return (inAgentHostNameStr.upper(), inAgentUserNameStr.upper(), inProcessNameWOExeStr.upper()) in __Orchestrator__.GSettingsGet()["ManagersProcessDict"]


def ProcessGet(inAgentHostNameStr: str, inAgentUserNameStr: str, inProcessNameWOExeStr: str) -> Process:
    """
    Return the process instance by the inProcessNameWOExeStr

    :param inAgentHostNameStr: Agent hostname in any case. Required to identify Process
    :param inAgentUserNameStr: Agent user name in any case. Required to identify Process
    :param inProcessNameWOExeStr: The process name without extension .exe (the key of the Process instance). Any case - will be processed to the upper case
    :return: Process instance (if exists) Else None
    """
    return __Orchestrator__.GSettingsGet()["ManagersProcessDict"].get((inAgentHostNameStr.upper(), inAgentUserNameStr.upper(), inProcessNameWOExeStr.upper()),None)

def ProcessStatusStrGet(inAgentHostNameStr: str, inAgentUserNameStr: str, inProcessNameWOExeStr: str) -> str:
    """
    Get the status of the Process instance.

    :param inAgentHostNameStr: Agent hostname in any case. Required to identify Process
    :param inAgentUserNameStr: Agent user name in any case. Required to identify Process
    :param inProcessNameWOExeStr: The process name without extension .exe (the key of the Process instance). Any case - will be processed to the upper case
    :return: Process status. See self.mStatusStr.
        Process instance has the following statuses:
        - 0_STOPPED
        - 1_STOPPED_MANUAL
        - 2_STOP_SAFE
        - 3_STOP_SAFE_MANUAL
        - 4_STARTED
        - 5_STARTED_MANUAL
        - None (if Process instance not exists)
    """
    lProcess = ProcessGet(inAgentHostNameStr = inAgentHostNameStr, inAgentUserNameStr = inAgentUserNameStr, inProcessNameWOExeStr=inProcessNameWOExeStr)
    if lProcess is not None: return lProcess.mStatusStr

def ProcessStart(inAgentHostNameStr: str, inAgentUserNameStr: str, inProcessNameWOExeStr: str, inIsManualBool: bool = True) -> str:
    """
    Manual/Auto start. Manual start will block scheduling execution. To return schedule execution use def Manual2Auto

    :param inAgentHostNameStr: Agent hostname in any case. Required to identify Process
    :param inAgentUserNameStr: Agent user name in any case. Required to identify Process
    :param inProcessNameWOExeStr: The process name without extension .exe (the key of the Process instance). Any case - will be processed to the upper case
    :param inIsManualBool: Default is True - Mark this operation as manual - StatusCheckStart/Stop will be blocked - only StatusCheck will be working. False - Auto operation
    :return: Process status. See self.mStatusStr.
        Process instance has the following statuses:
        - 0_STOPPED
        - 1_STOPPED_MANUAL
        - 2_STOP_SAFE
        - 3_STOP_SAFE_MANUAL
        - 4_STARTED
        - 5_STARTED_MANUAL
        - None (if Process instance not exists)
    """
    lProcess = ProcessGet(inAgentHostNameStr = inAgentHostNameStr, inAgentUserNameStr = inAgentUserNameStr, inProcessNameWOExeStr=inProcessNameWOExeStr)
    if lProcess is not None: return lProcess.Start(inIsManualBool=inIsManualBool)


def ProcessStopSafe(inAgentHostNameStr: str, inAgentUserNameStr: str, inProcessNameWOExeStr: str, inIsManualBool: bool = True, inStopSafeTimeoutSecFloat = None) -> str:
    """
    Manual/Auto stop safe. Stop safe is the operation which send signal to process to terminate own work (send term signal to process). Managers.Process wait for the mStopSafeTimeoutSecFloat seconds. After that, if process is not terminated - self will StopForce it.
    Manual stop safe will block scheduling execution. To return schedule execution use def Manual2Auto

    :param inAgentHostNameStr: Agent hostname in any case. Required to identify Process
    :param inAgentUserNameStr: Agent user name in any case. Required to identify Process
    :param inProcessNameWOExeStr: The process name without extension .exe (the key of the Process instance). Any case - will be processed to the upper case
    :param inIsManualBool: Default is True - Mark this operation as manual - StatusCheckStart/Stop will be blocked - only StatusCheck will be working. False - Auto operation
    :param inStopSafeTimeoutSecFloat: Default value goes from the instance. You can specify time is second to wait while safe stop. After that program will stop force
    :return: Process status. See self.mStatusStr.
        Process instance has the following statuses:
        - 0_STOPPED
        - 1_STOPPED_MANUAL
        - 2_STOP_SAFE
        - 3_STOP_SAFE_MANUAL
        - 4_STARTED
        - 5_STARTED_MANUAL
        - None (if Process instance not exists)
    """
    lProcess = ProcessGet(inAgentHostNameStr = inAgentHostNameStr, inAgentUserNameStr = inAgentUserNameStr, inProcessNameWOExeStr=inProcessNameWOExeStr)
    if lProcess is not None: return lProcess.StopSafe(inIsManualBool=inIsManualBool)

def ProcessStopForce(inAgentHostNameStr: str, inAgentUserNameStr: str, inProcessNameWOExeStr: str, inIsManualBool: bool = True) -> str:
    """
    Manual/Auto stop force. Force stop dont wait process termination  - it just terminate process now.
    Manual stop safe will block scheduling execution. To return schedule execution use def Manual2Auto

    :param inAgentHostNameStr: Agent hostname in any case. Required to identify Process
    :param inAgentUserNameStr: Agent user name in any case. Required to identify Process
    :param inProcessNameWOExeStr: The process name without extension .exe (the key of the Process instance). Any case - will be processed to the upper case
    :param inIsManualBool: Default is True - Mark this operation as manual - StatusCheckStart/Stop will be blocked - only StatusCheck will be working. False - Auto operation
    :return: Process status. See self.mStatusStr.
        Process instance has the following statuses:
        - 0_STOPPED
        - 1_STOPPED_MANUAL
        - 2_STOP_SAFE
        - 3_STOP_SAFE_MANUAL
        - 4_STARTED
        - 5_STARTED_MANUAL
        - None (if Process instance not exists)
    """
    lProcess = ProcessGet(inAgentHostNameStr = inAgentHostNameStr, inAgentUserNameStr = inAgentUserNameStr, inProcessNameWOExeStr=inProcessNameWOExeStr)
    if lProcess is not None: return lProcess.StopForce(inIsManualBool=inIsManualBool)

def ProcessStatusSave(inAgentHostNameStr: str, inAgentUserNameStr: str, inProcessNameWOExeStr: str):
    """
    Save current status of the process. After that you can restore process activity. Work when orchestrator is restarted. Don't save "STOP_SAFE" status > "STOPPED"

    :return: Process status. See self.mStatusStr. 0_STOPPED 1_STOPPED_MANUAL 2_STOP_SAFE 3_STOP_SAFE_MANUAL 4_STARTED 5_STARTED_MANUAL
    """
    lProcess = ProcessGet(inAgentHostNameStr=inAgentHostNameStr, inAgentUserNameStr=inAgentUserNameStr,
                          inProcessNameWOExeStr=inProcessNameWOExeStr)
    if lProcess is not None:
        lProcess.StatusSave()
        return lProcess.mStatusStr

def ProcessStatusRestore(inAgentHostNameStr: str, inAgentUserNameStr: str, inProcessNameWOExeStr: str):
    """
    Execute the StatusCheck, after that restore status to the saved state (see StatusSave). Work when orchestrator is restarted.

    :return: Process status. See self.mStatusStr. 0_STOPPED 1_STOPPED_MANUAL 2_STOP_SAFE 3_STOP_SAFE_MANUAL 4_STARTED 5_STARTED_MANUAL
    """
    lProcess = ProcessGet(inAgentHostNameStr=inAgentHostNameStr, inAgentUserNameStr=inAgentUserNameStr,
                          inProcessNameWOExeStr=inProcessNameWOExeStr)
    if lProcess is not None:
        lProcess.StatusRestore()
        return lProcess.mStatusStr

def ProcessStatusCheck(inAgentHostNameStr: str, inAgentUserNameStr: str, inProcessNameWOExeStr: str) -> str:
    """
    Check if process is alive. The def will save the manual flag is exists.

    :param inAgentHostNameStr: Agent hostname in any case. Required to identify Process
    :param inAgentUserNameStr: Agent user name in any case. Required to identify Process
    :param inProcessNameWOExeStr: The process name without extension .exe (the key of the Process instance). Any case - will be processed to the upper case
    :return: Process status. See self.mStatusStr.
        Process instance has the following statuses:
        - 0_STOPPED
        - 1_STOPPED_MANUAL
        - 2_STOP_SAFE
        - 3_STOP_SAFE_MANUAL
        - 4_STARTED
        - 5_STARTED_MANUAL
        - None (if Process instance not exists)
    """
    lProcess = ProcessGet(inAgentHostNameStr=inAgentHostNameStr, inAgentUserNameStr=inAgentUserNameStr,
                          inProcessNameWOExeStr=inProcessNameWOExeStr)
    if lProcess is not None:
        lProcess.StatusCheck()
        return lProcess.mStatusStr

def ProcessManual2Auto(inAgentHostNameStr: str, inAgentUserNameStr: str, inProcessNameWOExeStr: str) -> str:
    """
    Remove Manual flag from process (if exists) - it will allow the schedule operations via def StatusCheckStart(self):    def StatusCheckStorForce(self): def StatusCheckStopSafe(self):

    :param inAgentHostNameStr: Agent hostname in any case. Required to identify Process
    :param inAgentUserNameStr: Agent user name in any case. Required to identify Process
    :param inProcessNameWOExeStr: The process name without extension .exe (the key of the Process instance). Any case - will be processed to the upper case
    :return: Process status. See self.mStatusStr.
        Process instance has the following statuses:
        - 0_STOPPED
        - 1_STOPPED_MANUAL
        - 2_STOP_SAFE
        - 3_STOP_SAFE_MANUAL
        - 4_STARTED
        - 5_STARTED_MANUAL
        - None (if Process instance not exists)
    """
    lProcess = ProcessGet(inAgentHostNameStr=inAgentHostNameStr, inAgentUserNameStr=inAgentUserNameStr,
                          inProcessNameWOExeStr=inProcessNameWOExeStr)
    if lProcess is not None: return lProcess.Manual2Auto()

def ProcessManualStopTriggerSet(inAgentHostNameStr: str, inAgentUserNameStr: str, inProcessNameWOExeStr: str,  inMSTdTSecFloat: float, inMSTdNInt: int) -> None:
    """
    Remove Manual flag from process (if exists) - it will allow the schedule operations via def StatusCheckStart(self):    def StatusCheckStorForce(self): def StatusCheckStopSafe(self):

    :param inAgentHostNameStr: Agent hostname in any case. Required to identify Process
    :param inAgentUserNameStr: Agent user name in any case. Required to identify Process
    :param inProcessNameWOExeStr: The process name without extension .exe (the key of the Process instance). Any case - will be processed to the upper case
    :param inMSTdTSecFloat: Time periods in seconds
    :param inMSTdNInt: Counts of the start tries
    :return: None
    """
    lProcess = ProcessGet(inAgentHostNameStr=inAgentHostNameStr, inAgentUserNameStr=inAgentUserNameStr,
                          inProcessNameWOExeStr=inProcessNameWOExeStr)
    if lProcess is not None: lProcess.ManualStopTriggerSet(inMSTdTSecFloat = inMSTdTSecFloat, inMSTdNInt = inMSTdNInt)


def ProcessManualStopListClear(inAgentHostNameStr: str, inAgentUserNameStr: str, inProcessNameWOExeStr: str) -> None:
    """
    Clear the last start tries list.

    :param inAgentHostNameStr: Agent hostname in any case. Required to identify Process
    :param inAgentUserNameStr: Agent user name in any case. Required to identify Process
    :param inProcessNameWOExeStr: The process name without extension .exe (the key of the Process instance). Any case - will be processed to the upper case
    :return: None
    """
    lProcess = ProcessGet(inAgentHostNameStr=inAgentHostNameStr, inAgentUserNameStr=inAgentUserNameStr,
                          inProcessNameWOExeStr=inProcessNameWOExeStr)
    if lProcess is not None: lProcess.ManualStopListClear()

def ProcessScheduleStatusCheckEverySeconds(inAgentHostNameStr: str, inAgentUserNameStr: str, inProcessNameWOExeStr: str,inIntervalSecondsInt: int = 120):
    """
    Run status check every interval in second you specify.

    :param inAgentHostNameStr: Agent hostname in any case. Required to identify Process
    :param inAgentUserNameStr: Agent user name in any case. Required to identify Process
    :param inProcessNameWOExeStr: The process name without extension .exe (the key of the Process instance). Any case - will be processed to the upper case
    :param inIntervalSecondsInt: Interval in seconds. Default is 120
    :return: None
    """
    lProcess = ProcessGet(inAgentHostNameStr=inAgentHostNameStr, inAgentUserNameStr=inAgentUserNameStr,
                          inProcessNameWOExeStr=inProcessNameWOExeStr)
    # Check job in threaded way
    __Orchestrator__.OrchestratorScheduleGet().every(inIntervalSecondsInt).seconds.do(__Orchestrator__.OrchestratorThreadStart,lProcess.StatusCheck)